from unittest import TestCase

from hackle.decision import DecisionReason
from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.evaluator.experiment.experiment_evaluation import ExperimentEvaluation
from hackle.internal.evaluation.evaluator.remoteconfig.remote_config_evaluation import RemoteConfigEvaluation
from hackle.internal.event.user_event import RemoteConfigEvent, ExposureEvent
from hackle.internal.event.user_event_factory import UserEventFactory
from hackle.internal.model.entities import ParameterConfiguration
from hackle.internal.model.properties_builder import PropertiesBuilder
from hackle.internal.time.clock import Clock
from tests.internal.evaluation.evaluator.remoteconfig.test_remote_config_request import create_remote_config_request
from tests.internal.model.test_entities import create_experiment


class UserEventFactoryTest(TestCase):
    def test_create(self):
        class ClockStub(Clock):
            def current_millis(self):
                return 47

        sut = UserEventFactory(ClockStub())

        context = Evaluator.context()
        evaluation_1 = ExperimentEvaluation(
            DecisionReason.TRAFFIC_ALLOCATED,
            [],
            create_experiment(id=1),
            42,
            'B',
            ParameterConfiguration(42, {})
        )

        evaluation_2 = ExperimentEvaluation(
            DecisionReason.DEFAULT_RULE,
            [],
            create_experiment(id=2, type='FEATURE_FLAG'),
            320,
            'A',
            None
        )

        context.add_evaluation(evaluation_1)
        context.add_evaluation(evaluation_2)

        request = create_remote_config_request()
        evaluation = RemoteConfigEvaluation.of(
            request,
            context,
            999,
            'RC',
            DecisionReason.TARGET_RULE_MATCH,
            PropertiesBuilder()
        )

        events = sut.create(request, evaluation)

        self.assertTrue(3, len(events))

        self.assertIsInstance(events[0], RemoteConfigEvent)
        self.assertEqual(47, events[0].timestamp)
        self.assertEqual(request.user, events[0].user)
        self.assertEqual(request.parameter, events[0].parameter)
        self.assertEqual(999, events[0].value_id)
        self.assertEqual(DecisionReason.TARGET_RULE_MATCH, events[0].reason)
        self.assertEqual({'returnValue': 'RC'}, events[0].properties)

        self.assertIsInstance(events[1], ExposureEvent)
        self.assertEqual(47, events[1].timestamp)
        self.assertEqual(request.user, events[1].user)
        self.assertEqual(evaluation_1.experiment, events[1].experiment)
        self.assertEqual(42, events[1].variation_id)
        self.assertEqual('B', events[1].variation_key)
        self.assertEqual(DecisionReason.TRAFFIC_ALLOCATED, events[1].reason)
        self.assertEqual(
            {
                '$targetingRootType': 'REMOTE_CONFIG',
                '$targetingRootId': 1,
                '$parameterConfigurationId': 42
            },
            events[1].properties)

        self.assertIsInstance(events[2], ExposureEvent)
        self.assertEqual(47, events[2].timestamp)
        self.assertEqual(request.user, events[2].user)
        self.assertEqual(evaluation_2.experiment, events[2].experiment)
        self.assertEqual(320, events[2].variation_id)
        self.assertEqual('A', events[2].variation_key)
        self.assertEqual(DecisionReason.DEFAULT_RULE, events[2].reason)
        self.assertEqual(
            {
                '$targetingRootType': 'REMOTE_CONFIG',
                '$targetingRootId': 1,
            },
            events[2].properties)



import time
from unittest import TestCase
from unittest.mock import Mock

from hackle.internal.concurrent.schedule.scheduler import Scheduler, ScheduledJob
from hackle.internal.concurrent.schedule.thread_scheduler import ThreadScheduler
from hackle.internal.http.http_client import HttpClient
from hackle.internal.workspace.workspace_fetcher import WorkspaceFetcher


class WorkspaceFetcherTest(TestCase):

    def setUp(self):
        self.http_client = Mock(spec=HttpClient)
        self.http_client.get.return_value = '{}'

    def __workspace_fetcher(self, http_client=None, polling_interval_seconds=10.0, scheduler=ThreadScheduler()):
        return WorkspaceFetcher('localhost', http_client or self.http_client, polling_interval_seconds, scheduler)

    def test__fetch__without_start(self):
        sut = self.__workspace_fetcher()
        self.assertIsNone(sut.fetch())

    def test__start__poll(self):
        sut = self.__workspace_fetcher()

        sut.start()

        self.assertIsNotNone(sut.fetch())

    def test__start__once(self):
        scheduler = Mock(spec=Scheduler)
        scheduler.schedule_periodically.return_value = Mock()

        sut = self.__workspace_fetcher(scheduler=scheduler)

        for _ in range(0, 10):
            sut.start()

        scheduler.schedule_periodically.assert_called_once()

    def test__start__long_task_polling(self):
        http_client = DelayedHttpClient(0.25)
        sut = self.__workspace_fetcher(http_client=http_client, polling_interval_seconds=0.1)

        sut.start()
        time.sleep(0.4)

        self.assertEqual(3, http_client.call_count)  # 0, 250, 350 fetch / 100, 200 ignored

    def test__stop__cancel_polling_job(self):
        scheduler = Mock(spec=Scheduler)
        job = Mock(spec=ScheduledJob)
        scheduler.schedule_periodically.return_value = job

        sut = self.__workspace_fetcher(scheduler=scheduler)
        sut.start()
        sut.stop()

        job.cancel.assert_called_once()

    def test__stop__without_start(self):
        scheduler = Mock(spec=Scheduler)
        job = Mock(spec=ScheduledJob)
        scheduler.schedule_periodically.return_value = job

        sut = self.__workspace_fetcher(scheduler=scheduler)
        sut.stop()

        job.cancel.assert_not_called()


class DelayedHttpClient(object):

    def __init__(self, delay):
        self.__delay = delay
        self.__count = 0

    def get(self, url):
        self.__count += 1
        time.sleep(self.__delay)
        return '{}'

    @property
    def call_count(self):
        return self.__count



import abc

from six import add_metaclass

from hackle.internal.model.entities import Experiment
from hackle.internal.user.internal_hackle_user import InternalHackleUser
from hackle.internal.workspace.workspace import Workspace


@add_metaclass(abc.ABCMeta)
class Evaluator(object):

    @abc.abstractmethod
    def evaluate(self, request, context):
        """
        :param Evaluator.Request request:
        :param Evaluator.Context context:

        :rtype: Evaluator.Evaluation
        """
        pass

    class Key(object):
        def __init__(self, type_, id_):
            """
            :param str type_:
            :param int id_:
            """
            self.__type = type_
            self.__id = id_

        @property
        def type(self):
            """
            :rtype: str
            """
            return self.__type

        @property
        def id(self):
            """
            :rtype: int
            """
            return self.__id

        def __eq__(self, other):
            if not isinstance(other, Evaluator.Key):
                return False
            return self.__type == other.__type and self.__id == other.__id

        def __ne__(self, other):
            return not self.__eq__(other)

        def __str__(self):
            return "Evaluator.Key(type={}, id={})".format(self.__type, self.__id)

        def __repr__(self):
            return self.__str__()

    @add_metaclass(abc.ABCMeta)
    class Request(object):

        def __init__(self, key, workspace, user):
            """
            :param Evaluator.Key key:
            :param Workspace workspace:
            :param InternalHackleUser user:
            """
            self.__key = key
            self.__workspace = workspace
            self.__user = user

        @property
        def key(self):
            """
            :rtype: Evaluator.Key
            """
            return self.__key

        @property
        def workspace(self):
            """
            :rtype: Workspace
            """
            return self.__workspace

        @property
        def user(self):
            """
            :rtype: InternalHackleUser
            """
            return self.__user

        def __eq__(self, other):
            if not isinstance(other, Evaluator.Request):
                return False
            return self.__key == other.__key

        def __ne__(self, other):
            return not self.__eq__(other)

        def __str__(self):
            return 'Evaluator.Request({}, {})'.format(self.__key.type, self.__key.id)

        def __repr__(self):
            return self.__str__()

    @add_metaclass(abc.ABCMeta)
    class Evaluation(object):
        def __init__(self, reason, target_evaluations):
            """
            :param str reason:
            :param list[Evaluator.Evaluation] target_evaluations:
            """
            self.__reason = reason
            self.__target_evaluations = target_evaluations

        @property
        def reason(self):
            """
            :rtype: str
            """
            return self.__reason

        @property
        def target_evaluations(self):
            """
            :rtype: list[Evaluator.Evaluation]
            """
            return self.__target_evaluations

    class Context(object):
        def __init__(self):
            self.__stack = []
            self.__evaluations = []

        @property
        def stack(self):
            """
            :rtype: list[Evaluator.Request]
            """
            return list(self.__stack)

        @property
        def target_evaluations(self):
            """
            :rtype: list[Evaluator.Evaluation]
            """
            return list(self.__evaluations)

        def contains_request(self, request):
            """
            :param Evaluator.Request request:

            :rtype: bool
            """
            return request in self.__stack

        def add_request(self, request):
            """
            :param Evaluator.Request request:
            """
            self.__stack.append(request)

        @abc.abstractmethod
        def remove_request(self, request):
            """
            :param Evaluator.Request request:
            """
            self.__stack.remove(request)

        @abc.abstractmethod
        def get_evaluation_or_none(self, experiment):
            """
            :param Experiment experiment:
            :rtype: Evaluator.Evaluation or None
            """
            for evaluation in self.__evaluations:
                from hackle.internal.evaluation.evaluator.experiment.experiment_evaluation import ExperimentEvaluation
                if isinstance(evaluation, ExperimentEvaluation) and evaluation.experiment.id == experiment.id:
                    return evaluation
            return None

        @abc.abstractmethod
        def add_evaluation(self, evaluation):
            """
            :param Evaluator.Evaluation evaluation:
            """
            self.__evaluations.append(evaluation)

    @staticmethod
    def context():
        """
        :rtype: Evaluator.Context
        """
        return Evaluator.Context()


@add_metaclass(abc.ABCMeta)
class ContextualEvaluator(Evaluator):

    @abc.abstractmethod
    def supports(self, request):
        """
        :param Evaluator.Request request:

        :rtype: bool
        """
        pass

    @abc.abstractmethod
    def evaluate_internal(self, request, context):
        """
        :param Evaluator.Request request:
        :param Evaluator.Context context:

        :rtype: Evaluator.Evaluation
        """
        pass

    def evaluate(self, request, context):
        if context.contains_request(request):
            stack = context.stack + [request]
            raise Exception('Circular evaluation has occurred [{}]'.format(' - '.join(map(str, stack))))
        context.add_request(request)
        try:
            return self.evaluate_internal(request, context)
        finally:
            context.remove_request(request)



from unittest import TestCase
from unittest.mock import Mock

from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.match.target_matcher import TargetMatcher
from hackle.internal.evaluation.target.experiment_target_rule_determiner import ExperimentTargetRuleDeterminer
from hackle.internal.model.entities import *
from tests.internal.evaluation.evaluator.experiment.test_experiment_request import create_experiment_request
from tests.internal.model.test_entities import create_experiment


class ExperimentTargetRuleDeterminerTest(TestCase):

    def setUp(self):
        self.target_matches = []
        self.target_matcher = Mock(spec=TargetMatcher)
        self.target_matcher.matches = Mock(side_effect=self.target_matches)
        self.sut = ExperimentTargetRuleDeterminer(self.target_matcher)

    def test_등록된_rule이_없는경우_None리턴(self):
        # given
        experiment = create_experiment(target_rules=[])
        request = create_experiment_request(experiment=experiment)

        # when
        actual = self.sut.determine_target_rule_or_none(request, Evaluator.context())

        # then
        self.assertIsNone(actual)

    def test_등록된_규칙중_일치하는_첫번쨰_규칙을_리턴한다(self):
        # given
        experiment = create_experiment(target_rules=[
            self._rule(False),
            self._rule(False),
            self._rule(False),
            self._rule(False),
            self._rule(True),
            self._rule(False),
        ])
        request = create_experiment_request(experiment=experiment)

        # when
        actual = self.sut.determine_target_rule_or_none(request, Evaluator.context())

        # then
        self.assertEqual(experiment.target_rules[4], actual)
        self.assertEqual(5, self.target_matcher.matches.call_count)

    def test_등록된_규칙중_일치하는_규칙이_하나도_없는경우_None리턴(self):
        # given

        experiment = create_experiment(target_rules=[
            self._rule(False),
            self._rule(False),
            self._rule(False),
            self._rule(False),
            self._rule(False),
            self._rule(False),
        ])
        request = create_experiment_request(experiment=experiment)

        # when
        actual = self.sut.determine_target_rule_or_none(request, Evaluator.context())

        # then
        self.assertIsNone(actual)
        self.assertEqual(6, self.target_matcher.matches.call_count)

    def _rule(self, is_match):
        self.target_matches.append(is_match)
        target_rule = Mock(spec=TargetRule)
        target_rule.target = Mock()
        return target_rule



from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.evaluator.experiment.experiment_request import ExperimentRequest
from hackle.internal.model.entities import Experiment, ParameterConfiguration, Variation


class ExperimentEvaluation(Evaluator.Evaluation):
    def __init__(self, reason, target_evaluations, experiment, variation_id, variation_key, config):
        """
        :param str reason:
        :param list[Evaluator.Evaluation] target_evaluations:
        :param Experiment experiment:
        :param int or None variation_id:
        :param str variation_key:
        :param ParameterConfiguration or None config:
        """
        super(ExperimentEvaluation, self).__init__(reason, target_evaluations)
        self.__experiment = experiment
        self.__variation_id = variation_id
        self.__variation_key = variation_key
        self.__config = config

    @staticmethod
    def of(request, context, variation, reason):
        """
        :param ExperimentRequest request:
        :param Evaluator.Context context:
        :param Variation variation:
        :param str reason:

        :rtype: ExperimentEvaluation
        """
        parameter_configuration = None
        parameter_configuration_id = variation.parameter_configuration_id
        if parameter_configuration_id is not None:
            parameter_configuration = request.workspace.get_parameter_configuration_or_none(parameter_configuration_id)
            if parameter_configuration is None:
                raise Exception('ParameterConfiguration[{}]'.format(parameter_configuration_id))

        return ExperimentEvaluation(
            reason,
            context.target_evaluations,
            request.experiment,
            variation.id,
            variation.key,
            parameter_configuration
        )

    @staticmethod
    def of_default(request, context, reason):
        """
        :param ExperimentRequest request:
        :param Evaluator.Context context:
        :param str reason:

        :rtype: ExperimentEvaluation
        """
        variation = request.experiment.get_variation_by_key_or_none(request.default_variation_key)
        if variation is not None:
            return ExperimentEvaluation.of(request, context, variation, reason)
        else:
            return ExperimentEvaluation(
                reason,
                context.target_evaluations,
                request.experiment,
                None,
                request.default_variation_key,
                None
            )

    @property
    def experiment(self):
        """
        :rtype: Experiment
        """
        return self.__experiment

    @property
    def variation_id(self):
        """
        :rtype: int or None
        """
        return self.__variation_id

    @property
    def variation_key(self):
        """
        :rtype: str
        """
        return self.__variation_key

    @property
    def config(self):
        """
        :rtype: ParameterConfiguration or None
        """
        return self.__config

    def copy_with(self, reason):
        """
        :param str reason:
        :rtype: ExperimentEvaluation
        """
        return ExperimentEvaluation(
            reason,
            self.target_evaluations,
            self.experiment,
            self.variation_id,
            self.variation_key,
            self.config
        )



from unittest import TestCase

from hackle.internal.model.sdk import Sdk


class SdkTest(TestCase):

    def test_sdk(self):
        sdk = Sdk('key', 'name', 'version')
        self.assertEqual('key', sdk.key)
        self.assertEqual('name', sdk.name)
        self.assertEqual('version', sdk.version)



import time
from unittest import TestCase
from unittest.mock import Mock

from six.moves.queue import Queue

from hackle.internal.concurrent.schedule.thread_scheduler import ThreadScheduler
from hackle.internal.event.event_dispatcher import EventDispatcher
from hackle.internal.event.event_processor import EventProcessor
from hackle.internal.event.user_event import UserEvent
from tests.internal.concurrent.test_concurrent_supports import Jobs


class EventProcessorTest(TestCase):

    def setUp(self):
        self.event_dispatcher = Mock(spec=EventDispatcher)
        self.flush_scheduler = ThreadScheduler()
        self.queue = Queue(maxsize=1000)

    def __event_processor(
            self,
            queue=None,
            event_dispatcher=None,
            event_dispatch_size=100,
            flush_scheduler=ThreadScheduler(),
            flush_interval_seconds=10.0,
            shutdown_timeout_seconds=10.0):
        return EventProcessor(
            queue or self.queue,
            event_dispatcher or self.event_dispatcher,
            event_dispatch_size,
            flush_scheduler,
            flush_interval_seconds,
            shutdown_timeout_seconds
        )

    def test__process__invalid_event(self):
        sut = self.__event_processor()
        sut.process('test')

        self.assertEqual(0, self.queue.qsize())

    def test__process__not_dispatched_if_batch_size_is_not_reached(self):
        sut = self.__event_processor(event_dispatch_size=2, flush_interval_seconds=10)
        sut.start()

        sut.process(Mock(spec=UserEvent))
        time.sleep(1)

        self.event_dispatcher.dispatch.assert_not_called()

    def test__process__dispatch_when_batch_size_is_reached(self):
        sut = self.__event_processor(event_dispatch_size=2, flush_interval_seconds=10)

        sut.start()

        event = Mock(spec=UserEvent)
        sut.process(event)
        time.sleep(0.1)

        self.event_dispatcher.dispatch.assert_not_called()

        sut.process(event)
        time.sleep(0.1)

        self.event_dispatcher.dispatch.assert_called_once()

        sut.stop()

    def test__process__dispatch_when_interval_is_reached(self):
        sut = self.__event_processor(event_dispatch_size=1000, flush_interval_seconds=0.5)
        sut.start()

        sut.process(Mock(spec=UserEvent))
        sut.process(Mock(spec=UserEvent))
        sut.process(Mock(spec=UserEvent))
        sut.process(Mock(spec=UserEvent))
        sut.process(Mock(spec=UserEvent))
        time.sleep(0.7)

        self.event_dispatcher.dispatch.assert_called_once()

    def test__process__not_dispatch_if_events_is_empty(self):
        sut = self.__event_processor(event_dispatch_size=1000, flush_interval_seconds=0.1)
        sut.start()

        time.sleep(1)

        self.event_dispatcher.dispatch.assert_not_called()

    def test__process__ignore_when_queue_is_full(self):
        q = Queue(maxsize=1)
        sut = self.__event_processor(queue=q, event_dispatch_size=1000, flush_interval_seconds=0.1)

        for _ in range(0, 10):
            sut.process(Mock(spec=UserEvent))

        self.assertEqual(1, q.qsize())

    def test__process__concurrency(self):

        sut = self.__event_processor(queue=Queue(maxsize=16 * 10_000),
                                     event_dispatch_size=10,
                                     flush_interval_seconds=0.01)
        event = Mock(spec=UserEvent)

        def task():
            for i in range(10_000):
                sut.process(event)

        jobs = Jobs(16, task)

        sut.start()
        jobs.start_and_wait()

        sut.stop()

        count = sum(map(lambda it: len(it[0][0]), self.event_dispatcher.dispatch.call_args_list))
        self.assertEqual(16 * 10_000, count)



from hackle.internal.model.properties_builder import PropertiesBuilder
from hackle.model import User, HackleUser
from hackle.internal.user.identifier_type import IdentifierType
from hackle.internal.user.identifiers_builder import IdentifiersBuilder
from hackle.internal.user.internal_hackle_user import InternalHackleUser


class HackleUserResolver(object):

    # noinspection PyMethodMayBeStatic
    def resolve_or_none(self, user):
        """
        :param HackleUser user:
         
        :rtype: InternalHackleUser or None 
        """
        if user is None:
            return None

        if isinstance(user, User):
            hackle_user = HackleUser.of(user)
        elif isinstance(user, HackleUser):
            hackle_user = user
        else:
            return None

        user_properties = PropertiesBuilder.sanitize(hackle_user.properties)

        identifiers_builder = IdentifiersBuilder()
        identifiers_builder.add_identifiers(hackle_user.identifiers)

        if hackle_user.id is not None:
            identifiers_builder.add(IdentifierType.ID, hackle_user.id)

        if hackle_user.user_id is not None:
            identifiers_builder.add(IdentifierType.USER, hackle_user.user_id)

        if hackle_user.device_id is not None:
            identifiers_builder.add(IdentifierType.DEVICE, hackle_user.device_id)

        identifiers = identifiers_builder.build()

        if len(identifiers) == 0:
            return None

        return InternalHackleUser(identifiers, user_properties)



import abc

from six import add_metaclass

from hackle.decision import DecisionReason
from hackle.internal.evaluation.action.action_resolver import ActionResolver
from hackle.internal.evaluation.container.container_resolver import ContainerResolver
from hackle.internal.evaluation.evaluator.experiment.experiment_evaluation import ExperimentEvaluation
from hackle.internal.evaluation.evaluator.experiment.experiment_request import ExperimentRequest
from hackle.internal.evaluation.flow.evaluation_flow import EvaluationFlow
from hackle.internal.evaluation.target.experiment_target_determiner import ExperimentTargetDeterminer
from hackle.internal.evaluation.target.experiment_target_rule_determiner import ExperimentTargetRuleDeterminer
from hackle.internal.evaluation.target.override_resolver import OverrideResolver


@add_metaclass(abc.ABCMeta)
class FlowEvaluator(object):

    @abc.abstractmethod
    def evaluate(self, request, context, next_flow):
        """
        :param ExperimentRequest request:
        :param Evaluator.Context context:
        :param EvaluationFlow next_flow:

        :rtype: ExperimentEvaluation
        """
        pass


class OverrideEvaluator(FlowEvaluator):

    def __init__(self, override_resolver):
        """
        :param OverrideResolver override_resolver:
        """
        self.__override_resolver = override_resolver

    def evaluate(self, request, context, next_flow):
        overridden_variation = self.__override_resolver.resolve_or_none(request, context)

        if overridden_variation is not None:
            if request.experiment.type == 'AB_TEST':
                return ExperimentEvaluation.of(request, context, overridden_variation, DecisionReason.OVERRIDDEN)
            elif request.experiment.type == 'FEATURE_FLAG':
                return ExperimentEvaluation.of(request, context, overridden_variation,
                                               DecisionReason.INDIVIDUAL_TARGET_MATCH)
            else:
                raise Exception('experiment type [{}]'.format(request.experiment.type))
        else:
            return next_flow.evaluate(request, context)


class DraftEvaluator(FlowEvaluator):
    def evaluate(self, request, context, next_flow):
        if request.experiment.status == 'DRAFT':
            return ExperimentEvaluation.of_default(request, context, DecisionReason.EXPERIMENT_DRAFT)
        else:
            return next_flow.evaluate(request, context)


class PausedEvaluator(FlowEvaluator):
    def evaluate(self, request, context, next_flow):
        if request.experiment.status == 'PAUSED':
            if request.experiment.type == 'AB_TEST':
                return ExperimentEvaluation.of_default(request, context, DecisionReason.EXPERIMENT_PAUSED)
            elif request.experiment.type == 'FEATURE_FLAG':
                return ExperimentEvaluation.of_default(request, context, DecisionReason.FEATURE_FLAG_INACTIVE)
            else:
                raise Exception('experiment type [{}]'.format(request.experiment.type))
        else:
            return next_flow.evaluate(request, context)


class CompletedEvaluator(FlowEvaluator):
    def evaluate(self, request, context, next_flow):
        if request.experiment.status == 'COMPLETED':
            winner_variation = request.experiment.winner_variation
            if winner_variation is None:
                raise Exception('Winner variation[{}]'.format(request.experiment.id))
            return ExperimentEvaluation.of(request, context, winner_variation, DecisionReason.EXPERIMENT_COMPLETED)
        else:
            return next_flow.evaluate(request, context)


class ExperimentTargetEvaluator(FlowEvaluator):

    def __init__(self, experiment_target_determiner):
        """
        :param ExperimentTargetDeterminer experiment_target_determiner:
        """
        self.__experiment_target_determiner = experiment_target_determiner

    def evaluate(self, request, context, next_flow):
        if request.experiment.type != 'AB_TEST':
            raise Exception('experiment type must be AB_TEST [{}]'.format(request.experiment.id))

        is_user_in_experiment_target = self.__experiment_target_determiner.is_user_in_experiment_target(request,
                                                                                                        context)
        if is_user_in_experiment_target:
            return next_flow.evaluate(request, context)
        else:
            return ExperimentEvaluation.of_default(request, context, DecisionReason.NOT_IN_EXPERIMENT_TARGET)


class TrafficAllocateEvaluator(FlowEvaluator):

    def __init__(self, action_resolver):
        """
        :param ActionResolver action_resolver:
        """
        self.__action_resolver = action_resolver

    def evaluate(self, request, context, next_flow):
        if request.experiment.status != 'RUNNING':
            raise Exception('experiment status must be RUNNING [{}]'.format(request.experiment.id))

        if request.experiment.type != 'AB_TEST':
            raise Exception('experiment type must be AB_TEST [{}]'.format(request.experiment.id))

        default_rule = request.experiment.default_rule
        variation = self.__action_resolver.resolve_or_none(request, default_rule)
        if not variation:
            return ExperimentEvaluation.of_default(request, context, DecisionReason.TRAFFIC_NOT_ALLOCATED)

        if variation.is_dropped:
            return ExperimentEvaluation.of_default(request, context, DecisionReason.VARIATION_DROPPED)

        return ExperimentEvaluation.of(request, context, variation, DecisionReason.TRAFFIC_ALLOCATED)


class TargetRuleEvaluator(FlowEvaluator):

    def __init__(self, target_rule_determiner, action_resolver):
        """
        :param ExperimentTargetRuleDeterminer target_rule_determiner:
        :param ActionResolver action_resolver:
        """
        self.__target_rule_determiner = target_rule_determiner
        self.__action_resolver = action_resolver

    def evaluate(self, request, context, next_flow):
        if request.experiment.status != 'RUNNING':
            raise Exception('experiment status must be RUNNING [{}]'.format(request.experiment.id))

        if request.experiment.type != 'FEATURE_FLAG':
            raise Exception('experiment type must be FEATURE_FLAG [{}]'.format(request.experiment.id))

        if request.user.identifiers.get(request.experiment.identifier_type) is None:
            return next_flow.evaluate(request, context)

        target_rule = self.__target_rule_determiner.determine_target_rule_or_none(request, context)
        if target_rule is None:
            return next_flow.evaluate(request, context)

        variation = self.__action_resolver.resolve_or_none(request, target_rule.action)
        if not variation:
            raise Exception('FeatureFlag must decide the variation [{}]'.format(request.experiment.id))

        return ExperimentEvaluation.of(request, context, variation, DecisionReason.TARGET_RULE_MATCH)


class DefaultRuleEvaluator(FlowEvaluator):

    def __init__(self, action_resolver):
        """
        :param ActionResolver action_resolver:
        """
        self.__action_resolver = action_resolver

    def evaluate(self, request, context, next_flow):
        if request.experiment.status != 'RUNNING':
            raise Exception('experiment status must be RUNNING [{}]'.format(request.experiment.id))

        if request.experiment.type != 'FEATURE_FLAG':
            raise Exception('experiment type must be FEATURE_FLAG [{}]'.format(request.experiment.id))

        if request.user.identifiers.get(request.experiment.identifier_type) is None:
            return ExperimentEvaluation.of_default(request, context, DecisionReason.DEFAULT_RULE)

        variation = self.__action_resolver.resolve_or_none(request, request.experiment.default_rule)
        if not variation:
            raise Exception('FeatureFlag must decide the variation [{}]'.format(request.experiment.id))

        return ExperimentEvaluation.of(request, context, variation, DecisionReason.DEFAULT_RULE)


class ContainerEvaluator(FlowEvaluator):

    def __init__(self, container_resolver):
        """
        :param ContainerResolver container_resolver:
        """
        self.__container_resolver = container_resolver

    def evaluate(self, request, context, next_flow):
        experiment = request.experiment
        container_id = experiment.container_id
        if container_id is None:
            return next_flow.evaluate(request, context)

        container = request.workspace.get_container_or_none(container_id)
        if container is None:
            raise Exception('Container[{}]'.format(container_id))

        if self.__container_resolver.is_user_in_container_group(request, container):
            return next_flow.evaluate(request, context)
        else:
            return ExperimentEvaluation.of_default(request, context, DecisionReason.NOT_IN_MUTUAL_EXCLUSION_EXPERIMENT)


class IdentifierEvaluator(FlowEvaluator):
    def evaluate(self, request, context, next_flow):
        if request.user.identifiers.get(request.experiment.identifier_type) is not None:
            return next_flow.evaluate(request, context)
        else:
            return ExperimentEvaluation.of_default(request, context, DecisionReason.IDENTIFIER_NOT_FOUND)



from hackle.internal.evaluation.action.action_resolver import ActionResolver
from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.evaluator.experiment.experiment_request import ExperimentRequest
from hackle.internal.evaluation.match.target_matcher import TargetMatcher
from hackle.internal.model.entities import Variation


class OverrideResolver(object):
    def __init__(self, target_matcher, action_resolver):
        """
        :param TargetMatcher target_matcher:
        :param ActionResolver action_resolver:
        """
        self.__target_matcher = target_matcher
        self.__action_resolver = action_resolver

    def resolve_or_none(self, request, context):
        """
        :param ExperimentRequest request:
        :param Evaluator.Context context:

        :rtype: Variation or None
        """
        user_overridden_variation = self.__resolve_user_override_or_none(request)
        if user_overridden_variation is not None:
            return user_overridden_variation

        return self.__resolve_segment_override(request, context)

    # noinspection PyMethodMayBeStatic
    def __resolve_user_override_or_none(self, request):
        """
        :param ExperimentRequest request:
        :rtype: Variation or None
        """
        experiment = request.experiment
        identifier = request.user.identifiers.get(experiment.identifier_type)
        if identifier is None:
            return None

        overridden_variation_id = experiment.user_overrides.get(identifier)
        if overridden_variation_id is None:
            return None

        return experiment.get_variation_by_id_or_none(overridden_variation_id)

    def __resolve_segment_override(self, request, context):
        """
        :param ExperimentRequest request:
        :param Evaluator.Context context:
        :rtype: Variation or None
        """
        for overridden_rule in request.experiment.segment_overrides:
            if self.__target_matcher.matches(request, context, overridden_rule.target):
                return self.__action_resolver.resolve_or_none(request, overridden_rule.action)

        return None



from hackle.internal.model.properties_builder import PropertiesBuilder
from hackle.internal.user.identifiers_builder import IdentifiersBuilder


class InternalHackleUser(object):
    def __init__(self, identifiers, properties):
        """

        :param dict[str, str] identifiers:
        :param dict[str, object] properties:
        """
        self.__identifiers = identifiers
        self.__properties = properties

    @staticmethod
    def builder():
        return InternalHackleUser.Builder()

    @property
    def identifiers(self):
        """
        :rtype: dict[str, str]
        """
        return self.__identifiers

    @property
    def properties(self):
        """
        :rtype: dict[str, object]
        """
        return self.__properties

    def __eq__(self, other):
        if not isinstance(other, InternalHackleUser):
            return False
        return self.identifiers == other.identifiers and self.properties == other.properties

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "HackleUser(identifiers={}, properties={})".format(self.__identifiers, self.__properties)

    def __repr__(self):
        return self.__str__()

    class Builder(object):

        def __init__(self):
            self.__identifiers = IdentifiersBuilder()
            self.__properties = PropertiesBuilder()

        def identifier(self, identifier_type, identifier_value):
            """
            :param str identifier_type:
            :param str identifier_value:
            :rtype: InternalHackleUser.Builder
            """
            self.__identifiers.add(identifier_type, identifier_value)
            return self

        def identifiers(self, identifiers):
            """
            :param dict[str, str] identifiers:
            :rtype: InternalHackleUser.Builder
            """
            self.__identifiers.add_identifiers(identifiers)
            return self

        def property(self, property_key, property_value):
            """
            :param str property_key:
            :param object property_value:
            :rtype: InternalHackleUser.Builder
            """
            self.__properties.add(property_key, property_value)
            return self

        def properties(self, properties):
            """
            :param dict[str, object] properties:
            :rtype: InternalHackleUser.Builder
            """
            self.__properties.add_properties(properties)
            return self

        def build(self):
            """
            :rtype: InternalHackleUser
            """
            return InternalHackleUser(self.__identifiers.build(), self.__properties.build())



from unittest import TestCase

from hackle.internal.evaluation.match.operator_matcher import *
from hackle.internal.evaluation.match.value_matcher import *
from hackle.internal.evaluation.match.value_operator_matcher import ValueOperatorMatcherFactory, ValueOperatorMatcher
from hackle.internal.model.entities import TargetMatch
from tests import base


class ValueOperatorMatcherTest(TestCase):

    def setUp(self):
        self.sut = ValueOperatorMatcher(ValueOperatorMatcherFactory())

    def test_match_values중_하나라도_일치하는_값이_있으면_true(self):
        # given
        match = TargetMatch(
            type='MATCH',
            operator='IN',
            value_type='NUMBER',
            values=[1, 2, 3]
        )

        # when
        actual = self.sut.matches(3, match)

        # then
        self.assertTrue(actual)

    def test_match_values중_일치하는_값이_하나도_없으면_false(self):
        # given
        match = TargetMatch(
            type='MATCH',
            operator='IN',
            value_type='NUMBER',
            values=[1, 2, 3]
        )

        # when
        actual = self.sut.matches(4, match)

        # then
        self.assertFalse(actual)

    def test_일치하는_값이_있지만_match_type이_NOT_MATCH면_false(self):
        # given
        match = TargetMatch(
            type='NOT_MATCH',
            operator='IN',
            value_type='NUMBER',
            values=[1, 2, 3]
        )

        # when
        actual = self.sut.matches(3, match)

        # then
        self.assertFalse(actual)

    def test_일치하는_값이_없지만_match_type이_NOT_MATCH면_true(self):
        # given
        match = TargetMatch(
            type='NOT_MATCH',
            operator='IN',
            value_type='NUMBER',
            values=[1, 2, 3]
        )

        # when
        actual = self.sut.matches(4, match)

        # then
        self.assertTrue(actual)

    def test_user_value_가_array_인_경우_하나라도_매칭되면_true(self):
        # given
        match = TargetMatch(
            type='MATCH',
            operator='IN',
            value_type='NUMBER',
            values=[1, 2, 3]
        )

        sut = ValueOperatorMatcher(ValueOperatorMatcherFactory())

        # when
        actual = sut.matches([-1, 0, 1], match)

        # then
        self.assertTrue(actual)

    def test_array_user_value_중_매치되는게_하나도_없으면_false(self):
        # given
        match = TargetMatch(
            type='MATCH',
            operator='IN',
            value_type='NUMBER',
            values=[1, 2, 3]
        )

        sut = ValueOperatorMatcher(ValueOperatorMatcherFactory())

        # when
        actual = sut.matches([4, 5, 6], match)

        # then
        self.assertFalse(actual)

    def test_array_user_value_가_empty_인_경우_false(self):
        # given
        match = TargetMatch(
            type='MATCH',
            operator='IN',
            value_type='NUMBER',
            values=[1, 2, 3]
        )

        sut = ValueOperatorMatcher(ValueOperatorMatcherFactory())

        # when
        actual = sut.matches([], match)

        # then
        self.assertFalse(actual)


class ValueOperatorMatcherFactoryTest(base.BaseTest):

    def test_value_matcher(self):
        sut = ValueOperatorMatcherFactory()
        self.assertIsInstance(sut.get_value_matcher_or_none('STRING'), StringValueMatcher)
        self.assertIsInstance(sut.get_value_matcher_or_none('NUMBER'), NumberValueMatcher)
        self.assertIsInstance(sut.get_value_matcher_or_none('BOOLEAN'), BoolValueMatcher)
        self.assertIsInstance(sut.get_value_matcher_or_none('VERSION'), VersionValueMatcher)

    def test_operator_matcher(self):
        sut = ValueOperatorMatcherFactory()
        self.assertIsInstance(sut.get_operator_matcher_or_none('IN'), InMatcher)
        self.assertIsInstance(sut.get_operator_matcher_or_none('CONTAINS'), ContainsMatcher)
        self.assertIsInstance(sut.get_operator_matcher_or_none('STARTS_WITH'), StartsWithMatcher)
        self.assertIsInstance(sut.get_operator_matcher_or_none('ENDS_WITH'), EndsWithMatcher)
        self.assertIsInstance(sut.get_operator_matcher_or_none('GT'), GreaterThanMatcher)
        self.assertIsInstance(sut.get_operator_matcher_or_none('GTE'), GreaterThanOrEqualToMatcher)
        self.assertIsInstance(sut.get_operator_matcher_or_none('LT'), LessThanMatcher)
        self.assertIsInstance(sut.get_operator_matcher_or_none('LTE'), LessThanOrEqualToMatcher)



from hackle.internal.evaluation.evaluator.experiment.experiment_request import ExperimentRequest
from hackle.internal.user.internal_hackle_user import InternalHackleUser
from hackle.internal.workspace.workspace import Workspace
from tests.internal.model.test_entities import create_experiment


def create_experiment_request(workspace=None, user=None, experiment=None, default_variation_key='A'):
    return ExperimentRequest.of(
        workspace=workspace or Workspace('{}'),
        user=user or InternalHackleUser(identifiers={'$id': 'user'}, properties={}),
        experiment=experiment or create_experiment(),
        default_variation_key=default_variation_key
    )



from hackle.internal.model.entities import Experiment, Variation, TargetAction, RemoteConfigParameter, \
    RemoteConfigParameterValue


def create_experiment(**params):
    """
    :rtype: Experiment
    """
    return Experiment(
        id=params.get("id", 1),
        key=params.get("key", 1),
        type=params.get("type", "AB_TEST"),
        identifier_type=params.get("identifier_type", '$id'),
        status=params.get('status', 'RUNNING'),
        version=params.get('version', 1),
        variations=params.get("variations", [create_variation(id=1, key="A"), create_variation(id=2, key="B")]),
        user_overrides=params.get("user_overrides", {}),
        segment_overrides=params.get('segment_overrides', []),
        target_audiences=params.get('target_audiences', []),
        target_rules=params.get('target_rules', []),
        default_rule=params.get('default_rule', TargetAction('BUCKET', bucket_id=1)),
        container_id=params.get('container_id'),
        winner_variation_id=params.get('winner_variation_id')
    )


def create_variation(**params):
    """
    :rtype: Variation
    """
    return Variation(
        id=params.get("id", 1),
        key=params.get("key", "A"),
        is_dropped=params.get("is_dropped", False),
        parameter_configuration_id=params.get("config_id")
    )


def create_parameter(**params):
    """
    :rtype: RemoteConfigParameter
    """
    return RemoteConfigParameter(
        id=params.get("id", 1),
        key=params.get('key', 'test_parameter_key'),
        type=params.get("type', 'STRING"),
        identifier_type=params.get("identifier_type", '$id'),
        target_rules=params.get('target_rules', []),
        default_value=params.get('default_value', RemoteConfigParameterValue(1, 'parameter_default'))
    )



from hackle.internal.http.http_client import HttpClient
from hackle.internal.logger.log import Log
from hackle.internal.time.time_unit import SECONDS
from hackle.internal.workspace.workspace import Workspace


class WorkspaceFetcher(object):
    def __init__(self, sdk_base_url, http_client, polling_interval_seconds, scheduler):
        """
        :param str sdk_base_url:
        :param HttpClient http_client:
        :param float polling_interval_seconds:
        :param Scheduler scheduler:
        """
        self.__sdk_url = sdk_base_url + '/api/v2/workspaces'
        self.__http_client = http_client
        self.__polling_interval_seconds = polling_interval_seconds
        self.__scheduler = scheduler
        self.__polling_job = None

        self.__workspace = None

    def fetch(self):
        """
        :rtype: Workspace or None
        """
        return self.__workspace

    def start(self):
        if self.__polling_job is None:
            self.__poll()
            self.__polling_job = self.__scheduler.schedule_periodically(
                self.__polling_interval_seconds,
                self.__polling_interval_seconds,
                SECONDS,
                self.__poll
            )

    def stop(self):
        if self.__polling_job is not None:
            self.__polling_job.cancel()
        self.__scheduler.close()

    def __poll(self):
        try:
            data = self.__http_client.get(self.__sdk_url)
            if data is not None:
                self.__workspace = Workspace(data)
        except Exception as e:
            Log.get().error('Failed to poll Workspace: {}'.format(str(e)))



from unittest import TestCase
from unittest.mock import Mock

from hackle.internal.evaluation.match.condition_matcher_factory import ConditionMatcherFactory
from hackle.internal.evaluation.match.experiment_condition_matcher import ExperimentConditionMatcher
from hackle.internal.evaluation.match.segment_condition_matcher import SegmentConditionMatcher
from hackle.internal.evaluation.match.user_condition_matcher import UserConditionMatcher


class ConditionMatcherFactoryTest(TestCase):

    def test_get_matcher(self):
        sut = ConditionMatcherFactory(Mock())
        self.assertIsInstance(sut.get_condition_matcher_or_none('USER_ID'), UserConditionMatcher)
        self.assertIsInstance(sut.get_condition_matcher_or_none('USER_PROPERTY'), UserConditionMatcher)
        self.assertIsInstance(sut.get_condition_matcher_or_none('HACKLE_PROPERTY'), UserConditionMatcher)
        self.assertIsInstance(sut.get_condition_matcher_or_none('SEGMENT'), SegmentConditionMatcher)
        self.assertIsInstance(sut.get_condition_matcher_or_none('AB_TEST'), ExperimentConditionMatcher)
        self.assertIsInstance(sut.get_condition_matcher_or_none('FEATURE_FLAG'), ExperimentConditionMatcher)
        self.assertIsNone(sut.get_condition_matcher_or_none('INVALID'))



from unittest import TestCase

from hackle.internal.evaluation.match.operator_matcher import *


class OperatorMatcherTest(TestCase):

    # IN
    def test_in_string(self):
        sut = InMatcher()
        self.assertTrue(sut.string_matches('abc', 'abc'))
        self.assertFalse(sut.string_matches('abc', 'abc1'))
        self.assertFalse(sut.string_matches('1', 1))
        self.assertFalse(sut.string_matches(1, 1))

    def test_in_number(self):
        sut = InMatcher()
        self.assertTrue(sut.number_matches(320, 320))
        self.assertFalse(sut.number_matches(320, 3200))
        self.assertFalse(sut.number_matches(320, '320'))
        self.assertFalse(sut.number_matches('320', '320'))
        self.assertFalse(sut.number_matches(0, False))
        self.assertFalse(sut.number_matches(False, 0))
        self.assertFalse(sut.number_matches(False, False))
        self.assertFalse(sut.number_matches(1, True))
        self.assertFalse(sut.number_matches(True, 1))
        self.assertFalse(sut.number_matches(True, True))

    def test_in_bool(self):
        sut = InMatcher()
        self.assertTrue(sut.bool_matches(True, True))
        self.assertFalse(sut.bool_matches(True, False))
        self.assertTrue(sut.bool_matches(False, False))
        self.assertFalse(sut.bool_matches(False, True))
        self.assertFalse(sut.bool_matches(True, 'True'))
        self.assertFalse(sut.bool_matches('True', True))
        self.assertFalse(sut.bool_matches('True', True))
        self.assertFalse(sut.number_matches(0, False))
        self.assertFalse(sut.number_matches(False, 0))
        self.assertFalse(sut.number_matches(False, False))
        self.assertFalse(sut.number_matches(1, True))
        self.assertFalse(sut.number_matches(True, 1))
        self.assertFalse(sut.number_matches(True, True))

    def test_in_version(self):
        sut = InMatcher()
        self.assertTrue(sut.version_matches(self.v('1.0.0'), self.v('1.0.0')))
        self.assertFalse(sut.version_matches(self.v('1.0.0'), self.v('2.0.0')))

    # CONTAINS
    def test_contains_string(self):
        sut = ContainsMatcher()
        self.assertTrue(sut.string_matches('abc', 'abc'))
        self.assertTrue(sut.string_matches('abc', 'a'))
        self.assertTrue(sut.string_matches('abc', 'b'))
        self.assertTrue(sut.string_matches('abc', 'c'))
        self.assertTrue(sut.string_matches('abc', 'ab'))
        self.assertFalse(sut.string_matches('abc', 'ac'))
        self.assertFalse(sut.string_matches('a', 'acb'))
        self.assertFalse(sut.string_matches('1', 1))
        self.assertFalse(sut.string_matches(1, 1))

    def test_continas_number(self):
        sut = ContainsMatcher()
        self.assertFalse(sut.number_matches(1, 1))
        self.assertFalse(sut.number_matches(11, 1))

    def test_contains_bool(self):
        sut = ContainsMatcher()
        self.assertFalse(sut.bool_matches(True, True))
        self.assertFalse(sut.bool_matches(True, False))
        self.assertFalse(sut.bool_matches(False, False))
        self.assertFalse(sut.bool_matches(False, True))

    def test_contains_version(self):
        sut = ContainsMatcher()
        self.assertFalse(sut.version_matches(self.v('1.0.0'), self.v('1.0.0')))
        self.assertFalse(sut.version_matches(self.v('1.0.0'), self.v('2.0.0')))

    # STARTS_WITH
    def test_starts_with_string(self):
        sut = StartsWithMatcher()
        self.assertTrue(sut.string_matches('abc', 'abc'))
        self.assertTrue(sut.string_matches('abc', 'a'))
        self.assertFalse(sut.string_matches('abc', 'b'))
        self.assertFalse(sut.string_matches('a', 'ab'))
        self.assertFalse(sut.string_matches('1', 1))
        self.assertFalse(sut.string_matches(1, 1))

    def test_starts_with_number(self):
        sut = StartsWithMatcher()
        self.assertFalse(sut.number_matches(1, 1))
        self.assertFalse(sut.number_matches(11, 1))
        self.assertFalse(sut.number_matches(1, 11))

    def test_starts_with_bool(self):
        sut = StartsWithMatcher()
        self.assertFalse(sut.bool_matches(True, True))
        self.assertFalse(sut.bool_matches(True, False))
        self.assertFalse(sut.bool_matches(False, False))
        self.assertFalse(sut.bool_matches(False, True))

    def test_starts_with_version(self):
        sut = StartsWithMatcher()
        self.assertFalse(sut.version_matches(self.v('1.0.0'), self.v('1.0.0')))
        self.assertFalse(sut.version_matches(self.v('1.0.0'), self.v('2.0.0')))

    # ENS_WITH
    def test_ends_with_string(self):
        sut = EndsWithMatcher()
        self.assertTrue(sut.string_matches('abc', 'abc'))
        self.assertFalse(sut.string_matches('abc', 'a'))
        self.assertTrue(sut.string_matches('abc', 'c'))
        self.assertTrue(sut.string_matches('abc', 'bc'))
        self.assertFalse(sut.string_matches('a', 'ab'))
        self.assertFalse(sut.string_matches('1', 1))
        self.assertFalse(sut.string_matches(1, 1))

    def test_ends_with_number(self):
        sut = EndsWithMatcher()
        self.assertFalse(sut.number_matches(1, 1))
        self.assertFalse(sut.number_matches(11, 1))
        self.assertFalse(sut.number_matches(1, 11))

    def test_ends_with_bool(self):
        sut = EndsWithMatcher()
        self.assertFalse(sut.bool_matches(True, True))
        self.assertFalse(sut.bool_matches(True, False))
        self.assertFalse(sut.bool_matches(False, False))
        self.assertFalse(sut.bool_matches(False, True))

    def test_ends_with_version(self):
        sut = EndsWithMatcher()
        self.assertFalse(sut.version_matches(self.v('1.0.0'), self.v('1.0.0')))
        self.assertFalse(sut.version_matches(self.v('1.0.0'), self.v('2.0.0')))

    # GT
    def test_gt_string(self):
        sut = GreaterThanMatcher()
        self.assertFalse(sut.string_matches('41', '42'))
        self.assertFalse(sut.string_matches('42', '42'))
        self.assertTrue(sut.string_matches('43', '42'))

        self.assertFalse(sut.string_matches('20230114', '20230115'))
        self.assertFalse(sut.string_matches('20230115', '20230115'))
        self.assertTrue(sut.string_matches('20230116', '20230115'))

        self.assertFalse(sut.string_matches('2023-01-14', '2023-01-15'))
        self.assertFalse(sut.string_matches('2023-01-15', '2023-01-15'))
        self.assertTrue(sut.string_matches('2023-01-16', '2023-01-15'))

        self.assertFalse(sut.string_matches('a', 'a'))
        self.assertTrue(sut.string_matches('a', 'A'))
        self.assertFalse(sut.string_matches('A', 'a'))
        self.assertTrue(sut.string_matches('aa', 'a'))
        self.assertFalse(sut.string_matches('a', 'aa'))

    def test_gt_number(self):
        sut = GreaterThanMatcher()
        self.assertTrue(sut.number_matches(1.0001, 1))
        self.assertTrue(sut.number_matches(2, 1))
        self.assertFalse(sut.number_matches(1, 1))
        self.assertFalse(sut.number_matches(1, 2))
        self.assertFalse(sut.number_matches(0.9999, 1))
        self.assertFalse(sut.number_matches('2', 1))
        self.assertFalse(sut.number_matches(2, '1'))

    def test_gt_bool(self):
        sut = GreaterThanMatcher()
        self.assertFalse(sut.bool_matches(True, True))
        self.assertFalse(sut.bool_matches(True, False))
        self.assertFalse(sut.bool_matches(False, False))
        self.assertFalse(sut.bool_matches(False, True))

    def test_gt_version(self):
        sut = GreaterThanMatcher()
        self.assertFalse(sut.version_matches(self.v('1.0.0'), self.v('1.0.0')))
        self.assertFalse(sut.version_matches(self.v('1.0.0'), self.v('2.0.0')))
        self.assertTrue(sut.version_matches(self.v('2.0.0'), self.v('1.0.0')))

    # GTE
    def test_gte_string(self):
        sut = GreaterThanOrEqualToMatcher()
        self.assertFalse(sut.string_matches('41', '42'))
        self.assertTrue(sut.string_matches('42', '42'))
        self.assertTrue(sut.string_matches('43', '42'))

        self.assertFalse(sut.string_matches('20230114', '20230115'))
        self.assertTrue(sut.string_matches('20230115', '20230115'))
        self.assertTrue(sut.string_matches('20230116', '20230115'))

        self.assertFalse(sut.string_matches('2023-01-14', '2023-01-15'))
        self.assertTrue(sut.string_matches('2023-01-15', '2023-01-15'))
        self.assertTrue(sut.string_matches('2023-01-16', '2023-01-15'))

        self.assertTrue(sut.string_matches('a', 'a'))
        self.assertTrue(sut.string_matches('a', 'A'))
        self.assertFalse(sut.string_matches('A', 'a'))
        self.assertTrue(sut.string_matches('aa', 'a'))
        self.assertFalse(sut.string_matches('a', 'aa'))

    def test_gte_number(self):
        sut = GreaterThanOrEqualToMatcher()
        self.assertTrue(sut.number_matches(1.0001, 1))
        self.assertTrue(sut.number_matches(2, 1))
        self.assertTrue(sut.number_matches(1, 1))
        self.assertFalse(sut.number_matches(1, 2))
        self.assertFalse(sut.number_matches(0.9999, 1))
        self.assertFalse(sut.number_matches('1', 1))
        self.assertFalse(sut.number_matches(1, '1'))

    def test_gte_bool(self):
        sut = GreaterThanOrEqualToMatcher()
        self.assertFalse(sut.bool_matches(True, True))
        self.assertFalse(sut.bool_matches(True, False))
        self.assertFalse(sut.bool_matches(False, False))
        self.assertFalse(sut.bool_matches(False, True))

    def test_gte_version(self):
        sut = GreaterThanOrEqualToMatcher()
        self.assertTrue(sut.version_matches(self.v('1.0.0'), self.v('1.0.0')))
        self.assertFalse(sut.version_matches(self.v('1.0.0'), self.v('2.0.0')))
        self.assertTrue(sut.version_matches(self.v('2.0.0'), self.v('1.0.0')))

    # LT
    def test_lt_string(self):
        sut = LessThanMatcher()
        self.assertTrue(sut.string_matches('41', '42'))
        self.assertFalse(sut.string_matches('42', '42'))
        self.assertFalse(sut.string_matches('43', '42'))

        self.assertTrue(sut.string_matches('20230114', '20230115'))
        self.assertFalse(sut.string_matches('20230115', '20230115'))
        self.assertFalse(sut.string_matches('20230116', '20230115'))

        self.assertTrue(sut.string_matches('2023-01-14', '2023-01-15'))
        self.assertFalse(sut.string_matches('2023-01-15', '2023-01-15'))
        self.assertFalse(sut.string_matches('2023-01-16', '2023-01-15'))

        self.assertFalse(sut.string_matches('a', 'a'))
        self.assertFalse(sut.string_matches('a', 'A'))
        self.assertTrue(sut.string_matches('A', 'a'))
        self.assertFalse(sut.string_matches('aa', 'a'))
        self.assertTrue(sut.string_matches('a', 'aa'))

    def test_lt_number(self):
        sut = LessThanMatcher()
        self.assertFalse(sut.number_matches(1.0001, 1))
        self.assertFalse(sut.number_matches(2, 1))
        self.assertFalse(sut.number_matches(1, 1))
        self.assertTrue(sut.number_matches(1, 2))
        self.assertTrue(sut.number_matches(0.9999, 1))
        self.assertFalse(sut.number_matches('1', 2))
        self.assertFalse(sut.number_matches(1, '2'))

    def test_lt_bool(self):
        sut = LessThanMatcher()
        self.assertFalse(sut.bool_matches(True, True))
        self.assertFalse(sut.bool_matches(True, False))
        self.assertFalse(sut.bool_matches(False, False))
        self.assertFalse(sut.bool_matches(False, True))

    def test_lt_version(self):
        sut = LessThanMatcher()
        self.assertFalse(sut.version_matches(self.v('1.0.0'), self.v('1.0.0')))
        self.assertTrue(sut.version_matches(self.v('1.0.0'), self.v('2.0.0')))
        self.assertFalse(sut.version_matches(self.v('2.0.0'), self.v('1.0.0')))

    # LTE
    def test_lte_string(self):
        sut = LessThanOrEqualToMatcher()
        self.assertTrue(sut.string_matches('41', '42'))
        self.assertTrue(sut.string_matches('42', '42'))
        self.assertFalse(sut.string_matches('43', '42'))

        self.assertTrue(sut.string_matches('20230114', '20230115'))
        self.assertTrue(sut.string_matches('20230115', '20230115'))
        self.assertFalse(sut.string_matches('20230116', '20230115'))

        self.assertTrue(sut.string_matches('2023-01-14', '2023-01-15'))
        self.assertTrue(sut.string_matches('2023-01-15', '2023-01-15'))
        self.assertFalse(sut.string_matches('2023-01-16', '2023-01-15'))

        self.assertTrue(sut.string_matches('a', 'a'))
        self.assertFalse(sut.string_matches('a', 'A'))
        self.assertTrue(sut.string_matches('A', 'a'))
        self.assertFalse(sut.string_matches('aa', 'a'))
        self.assertTrue(sut.string_matches('a', 'aa'))

    def test_lte_number(self):
        sut = LessThanOrEqualToMatcher()
        self.assertFalse(sut.number_matches(1.0001, 1))
        self.assertFalse(sut.number_matches(2, 1))
        self.assertTrue(sut.number_matches(1, 1))
        self.assertTrue(sut.number_matches(1, 2))
        self.assertTrue(sut.number_matches(0.9999, 1))
        self.assertFalse(sut.number_matches('1', 1))
        self.assertFalse(sut.number_matches(1, '1'))

    def test_lte_bool(self):
        sut = LessThanOrEqualToMatcher()
        self.assertFalse(sut.bool_matches(True, True))
        self.assertFalse(sut.bool_matches(True, False))
        self.assertFalse(sut.bool_matches(False, False))
        self.assertFalse(sut.bool_matches(False, True))

    def test_lte_version(self):
        sut = LessThanOrEqualToMatcher()
        self.assertTrue(sut.version_matches(self.v('1.0.0'), self.v('1.0.0')))
        self.assertTrue(sut.version_matches(self.v('1.0.0'), self.v('2.0.0')))
        self.assertFalse(sut.version_matches(self.v('2.0.0'), self.v('1.0.0')))

    @staticmethod
    def v(version):
        return SemanticVersion.parse_or_none(version)



from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.match.condition_matcher import ConditionMatcher
from hackle.internal.evaluation.match.experiment_condition_matcher import ExperimentConditionMatcher, \
    AbTestConditionMatcher, \
    FeatureFlagConditionMatcher
from hackle.internal.evaluation.match.segment_condition_matcher import SegmentConditionMatcher, SegmentMatcher
from hackle.internal.evaluation.match.user_condition_matcher import UserConditionMatcher, UserValueResolver
from hackle.internal.evaluation.match.value_operator_matcher import ValueOperatorMatcher, ValueOperatorMatcherFactory


class ConditionMatcherFactory(object):

    def __init__(self, evaluator):
        """
        :param Evaluator evaluator:
        """
        value_operator_matcher = ValueOperatorMatcher(ValueOperatorMatcherFactory())
        self.__user_condition_matcher = UserConditionMatcher(UserValueResolver(), value_operator_matcher)
        self.__segment_condition_matcher = SegmentConditionMatcher(SegmentMatcher(self.__user_condition_matcher))
        self.__experiment_condition_matcher = ExperimentConditionMatcher(
            AbTestConditionMatcher(evaluator, value_operator_matcher),
            FeatureFlagConditionMatcher(evaluator, value_operator_matcher)
        )

    def get_condition_matcher_or_none(self, target_key_type):
        """
        :param str target_key_type:

        :rtype: ConditionMatcher or None
        """
        if target_key_type == 'USER_ID':
            return self.__user_condition_matcher
        elif target_key_type == 'USER_PROPERTY':
            return self.__user_condition_matcher
        elif target_key_type == 'HACKLE_PROPERTY':
            return self.__user_condition_matcher
        elif target_key_type == 'SEGMENT':
            return self.__segment_condition_matcher
        elif target_key_type == 'AB_TEST':
            return self.__experiment_condition_matcher
        elif target_key_type == 'FEATURE_FLAG':
            return self.__experiment_condition_matcher
        else:
            return None



from unittest import TestCase
from unittest.mock import Mock

from hackle.decision import DecisionReason
from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.evaluator.remoteconfig.remote_config_evaluation import RemoteConfigEvaluation
from hackle.internal.model.entities import RemoteConfigParameter
from hackle.internal.model.properties_builder import PropertiesBuilder
from tests.internal.evaluation.evaluator.remoteconfig.test_remote_config_request import create_remote_config_request


class RemoteConfigEvaluationTest(TestCase):

    def test_create(self):
        parameter = Mock(spec=RemoteConfigParameter)
        parameter.id = 1

        request = create_remote_config_request(parameter=parameter)
        context = Evaluator.context()
        context.add_evaluation(Mock(spec=Evaluator.Evaluation))

        evaluation = RemoteConfigEvaluation.of(request, context, 42, "go", DecisionReason.DEFAULT_RULE,
                                               PropertiesBuilder())

        self.assertEqual(DecisionReason.DEFAULT_RULE, evaluation.reason)
        self.assertEqual(1, len(evaluation.target_evaluations))
        self.assertEqual(parameter, evaluation.parameter)
        self.assertEqual("go", evaluation.properties.get("returnValue"))

    def test_create_by_default(self):
        parameter = Mock(spec=RemoteConfigParameter)
        parameter.id = 1

        request = create_remote_config_request(parameter=parameter, default_value="42")
        context = Evaluator.context()
        context.add_evaluation(Mock(spec=Evaluator.Evaluation))

        evaluation = RemoteConfigEvaluation.of_default(request, context, DecisionReason.DEFAULT_RULE,
                                                       PropertiesBuilder())

        self.assertEqual(DecisionReason.DEFAULT_RULE, evaluation.reason)
        self.assertEqual(1, len(evaluation.target_evaluations))
        self.assertEqual(parameter, evaluation.parameter)
        self.assertEqual("42", evaluation.properties.get("returnValue"))



from unittest import TestCase
from unittest.mock import Mock

from hackle.internal.evaluation.match.user_condition_matcher import UserValueResolver, UserConditionMatcher
from hackle.internal.evaluation.match.value_operator_matcher import ValueOperatorMatcher
from hackle.internal.model.entities import TargetCondition, TargetKey, TargetMatch


class UserConditionMatcherTest(TestCase):

    def setUp(self):
        self.user_value_resolver = Mock(spec=UserValueResolver)
        self.value_operator_matcher = Mock(spec=ValueOperatorMatcher)
        self.sut = UserConditionMatcher(self.user_value_resolver, self.value_operator_matcher)

    def test_Key에_해당하는_UserValue가_없으면_match_false(self):
        # given
        condition = TargetCondition(
            TargetKey('USER_PROPERTY', "age"),
            TargetMatch('MATCH', 'IN', 'NUMBER', [30])
        )
        self.user_value_resolver.resolve_or_none.return_value = None

        # when
        actual = self.sut.matches(Mock(), Mock(), condition)

        # then
        self.assertFalse(actual)

    def test_Key에_해당하는_UserValue로_매칭한다(self):
        # given
        condition = TargetCondition(
            TargetKey('USER_PROPERTY', "age"),
            TargetMatch('MATCH', 'IN', 'NUMBER', [30])
        )
        self.user_value_resolver.resolve_or_none.return_value = "test_user_value"
        self.value_operator_matcher.matches.return_value = True

        # when
        actual = self.sut.matches(Mock(), Mock(), condition)

        # then
        self.assertTrue(actual)

    def test_Key에_해당하는_UserValue로_매칭한다_2(self):
        # given
        condition = TargetCondition(
            TargetKey('USER_PROPERTY', "gender"),
            TargetMatch('MATCH', 'IN', 'NUMBER', [0])
        )
        self.user_value_resolver.resolve_or_none.return_value = 0
        self.value_operator_matcher.matches.return_value = True

        # when
        actual = self.sut.matches(Mock(), Mock(), condition)

        # then
        self.assertTrue(actual)

    def test_Key에_해당하는_UserValue로_매칭한다_3(self):
        # given
        condition = TargetCondition(
            TargetKey('USER_PROPERTY', "is_admin"),
            TargetMatch('MATCH', 'IN', 'BOOLEAN', [False])
        )
        self.user_value_resolver.resolve_or_none.return_value = False
        self.value_operator_matcher.matches.return_value = True

        # when
        actual = self.sut.matches(Mock(), Mock(), condition)

        # then
        self.assertTrue(actual)



from hackle.internal.evaluation.bucket.bucketer import Bucketer
from hackle.internal.evaluation.evaluator.experiment.experiment_request import ExperimentRequest
from hackle.internal.logger.log import Log
from hackle.internal.model.entities import TargetAction, Variation


class ActionResolver(object):
    def __init__(self, bucketer):
        """
        :param Bucketer bucketer:
        """
        self.__bucketer = bucketer

    def resolve_or_none(self, request, action):
        """
        :param ExperimentRequest request:
        :param TargetAction action:
        :rtype: Variation or None
        """
        if action.type == 'VARIATION':
            return self.__resolve_variation(request, action)
        elif action.type == 'BUCKET':
            return self.__resolve_bucket(request, action)
        else:
            Log.get().debug('Unsupported type[{}]. Please use the latest version of sdk.'.format(action.type))
            return None

    # noinspection PyMethodMayBeStatic
    def __resolve_variation(self, request, action):
        """
        :param ExperimentRequest request:
        :param TargetAction action:
        :rtype: Variation
        """
        variation = request.experiment.get_variation_by_id_or_none(action.variation_id)
        if variation is None:
            raise Exception('variation[{}]'.format(action.variation_id))
        return variation

    def __resolve_bucket(self, request, action):
        """
         :param ExperimentRequest request:
         :param TargetAction action:
         :rtype: Variation or None
         """
        bucket = request.workspace.get_bucket_or_none(action.bucket_id)
        if bucket is None:
            raise Exception('bucket[{}]'.format(action.bucket_id))

        identifier = request.user.identifiers.get(request.experiment.identifier_type)
        if identifier is None:
            return None

        allocated_slot = self.__bucketer.bucketing(bucket, identifier)
        if allocated_slot is None:
            return None

        return request.experiment.get_variation_by_id_or_none(allocated_slot.variation_id)



from unittest import TestCase
from unittest.mock import Mock

from hackle.decision import DecisionReason
from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.evaluator.experiment.experiment_request import ExperimentRequest
from hackle.internal.evaluation.evaluator.remoteconfig.remote_config_evaluator import RemoteConfigEvaluator
from hackle.internal.evaluation.evaluator.remoteconfig.remote_config_request import RemoteConfigRequest
from hackle.internal.evaluation.target.remote_config_target_rule_determiner import RemoteConfigTargetRuleDeterminer
from hackle.internal.model.entities import RemoteConfigParameterValue, RemoteConfigTargetRule
from tests.internal.evaluation.evaluator.remoteconfig.test_remote_config_request import create_remote_config_request
from tests.internal.model.test_entities import create_parameter


class RemoteConfigEvaluatorTest(TestCase):

    def setUp(self):
        self.target_rule_determiner = Mock(spec=RemoteConfigTargetRuleDeterminer)
        self.target_rule_determiner.determine_target_rule_or_none.return_value = None

        self.sut = RemoteConfigEvaluator(self.target_rule_determiner)

    def test_supports(self):
        self.assertFalse(self.sut.supports(Mock(spec=ExperimentRequest)))
        self.assertTrue(self.sut.supports(Mock(spec=RemoteConfigRequest)))

    def test__evaluate__circular_call(self):
        request = create_remote_config_request()
        context = Evaluator.context()
        context.add_request(request)

        with self.assertRaises(Exception) as actual:
            self.sut.evaluate(request, context)

        self.assertIn('Circular evaluation has occurred', str(actual.exception))

    def test__evaluate__identifier_not_found(self):
        # given
        parameter = create_parameter(
            type='STRING',
            identifier_type='custom_id',
            default_value=RemoteConfigParameterValue(43, 'parameter_default')
        )

        request = create_remote_config_request(parameter=parameter, default_value='sdk_default')

        # when
        actual = self.sut.evaluate(request, Evaluator.context())

        # then
        self.assertEqual(DecisionReason.IDENTIFIER_NOT_FOUND, actual.reason)
        self.assertEqual('sdk_default', actual.value)
        self.assertEqual(
            {'requestValueType': 'STRING', 'requestDefaultValue': 'sdk_default', 'returnValue': 'sdk_default'},
            actual.properties)

    def test__evaluate__target_rule_match(self):
        # given
        target_rule = RemoteConfigTargetRule(
            'target_rule_key',
            'target_rule_name',
            Mock(),
            42,
            RemoteConfigParameterValue(320, 'target_rule_value')
        )

        parameter = create_parameter(
            type='STRING',
            target_rules=[target_rule],
            default_value=RemoteConfigParameterValue(43, 'parameter_default')
        )

        self.target_rule_determiner.determine_target_rule_or_none.return_value = target_rule

        request = create_remote_config_request(parameter=parameter, default_value='sdk_default')

        # when
        actual = self.sut.evaluate(request, Evaluator.context())

        # then
        self.assertEqual(DecisionReason.TARGET_RULE_MATCH, actual.reason)
        self.assertEqual(320, actual.value_id)
        self.assertEqual('target_rule_value', actual.value)
        self.assertEqual(
            {
                'requestValueType': 'STRING',
                'requestDefaultValue': 'sdk_default',
                'returnValue': 'target_rule_value',
                'targetRuleKey': 'target_rule_key',
                'targetRuleName': 'target_rule_name'
            },
            actual.properties)

    def test__evaluate__default_rule(self):
        # given
        target_rule = RemoteConfigTargetRule(
            'target_rule_key',
            'target_rule_name',
            Mock(),
            42,
            RemoteConfigParameterValue(320, 'target_rule_value')
        )

        parameter = create_parameter(
            type='STRING',
            target_rules=[target_rule],
            default_value=RemoteConfigParameterValue(43, 'parameter_default')
        )

        self.target_rule_determiner.determine_target_rule_or_none.return_value = None

        request = create_remote_config_request(parameter=parameter, default_value='sdk_default')

        # when
        actual = self.sut.evaluate(request, Evaluator.context())

        # then
        self.assertEqual(DecisionReason.DEFAULT_RULE, actual.reason)
        self.assertEqual(43, actual.value_id)
        self.assertEqual('parameter_default', actual.value)
        self.assertEqual(
            {
                'requestValueType': 'STRING',
                'requestDefaultValue': 'sdk_default',
                'returnValue': 'parameter_default',
            },
            actual.properties)

    def test_type_match(self):
        self.__assert_type_match('STRING', 'match_string', 'default_string', True)
        self.__assert_type_match('STRING', '', 'default_string', True)
        self.__assert_type_match('STRING', 0, 'default_string', False)
        self.__assert_type_match('STRING', 1, 'default_string', False)
        self.__assert_type_match('STRING', False, 'default_string', False)
        self.__assert_type_match('STRING', True, 'default_string', False)

        self.__assert_type_match('NUMBER', 0, 999, True)
        self.__assert_type_match('NUMBER', 1, 999, True)
        self.__assert_type_match('NUMBER', -1, 999, True)
        self.__assert_type_match('NUMBER', 0.0, 999, True)
        self.__assert_type_match('NUMBER', 1.0, 999, True)
        self.__assert_type_match('NUMBER', -1.0, 999, True)
        self.__assert_type_match('NUMBER', 1.1, 999, True)
        self.__assert_type_match('NUMBER', '1', 999, False)
        self.__assert_type_match('NUMBER', '0', 999, False)
        self.__assert_type_match('NUMBER', True, 999, False)
        self.__assert_type_match('NUMBER', False, 999, False)

        self.__assert_type_match('BOOLEAN', True, False, True)
        self.__assert_type_match('BOOLEAN', False, True, True)
        self.__assert_type_match('BOOLEAN', 0, True, False)
        self.__assert_type_match('BOOLEAN', 1, False, False)

        self.__assert_type_match('NULL', 'match_string', None, True)
        self.__assert_type_match('NULL', 1, None, True)
        self.__assert_type_match('NULL', True, None, True)
        self.__assert_type_match('NULL', False, None, True)

        self.__assert_type_match('VERSION', '1.0.0', 'default', False)
        self.__assert_type_match('JSON', '{}', 'default', False)
        self.__assert_type_match('UNKNOWN', '{}', 'default', False)

    def __assert_type_match(self, required_type, match_value, default_value, is_match):
        parameter = create_parameter(
            type='STRING',
            default_value=RemoteConfigParameterValue(43, match_value)
        )
        request = create_remote_config_request(
            parameter=parameter, required_type=required_type, default_value=default_value)

        actual = self.sut.evaluate(request, Evaluator.context())

        if is_match:
            self.assertEqual(43, actual.value_id)
            self.assertEqual(match_value, actual.value)
            self.assertEqual(DecisionReason.DEFAULT_RULE, actual.reason)
        else:
            self.assertEqual(None, actual.value_id)
            self.assertEqual(default_value, actual.value)
            self.assertEqual(DecisionReason.TYPE_MISMATCH, actual.reason)



from hackle.internal.evaluation.match.operator_matcher import *
from hackle.internal.evaluation.match.value_matcher import *
from hackle.internal.logger.log import Log


class ValueOperatorMatcher(object):

    def __init__(self, factory):
        self.factory = factory

    def matches(self, user_value, match):
        value_matcher = self.factory.get_value_matcher_or_none(match.value_type)
        if not value_matcher:
            Log.get().debug('Unsupported type[{}]. Please use the latest version of sdk.'.format(match.value_type))
            return False

        operator_matcher = self.factory.get_operator_matcher_or_none(match.operator)
        if not operator_matcher:
            Log.get().debug('Unsupported type[{}]. Please use the latest version of sdk.'.format(match.operator))
            return False

        is_matched = self.__matches(user_value, match, value_matcher, operator_matcher)
        return self.__type_matches(match.type, is_matched)

    def __matches(self, user_value, match, value_matcher, operator_matcher):
        if isinstance(user_value, list):
            return self.__array_matches(user_value, match, value_matcher, operator_matcher)
        else:
            return self.__single_matches(user_value, match, value_matcher, operator_matcher)

    # noinspection PyMethodMayBeStatic
    def __single_matches(self, user_value, match, value_matcher, operator_matcher):
        for match_value in match.values:
            if value_matcher.matches(operator_matcher, user_value, match_value):
                return True
        return False

    def __array_matches(self, user_values, match, value_matcher, operator_matcher):
        for user_value in user_values:
            if self.__single_matches(user_value, match, value_matcher, operator_matcher):
                return True
        return False

    # noinspection PyMethodMayBeStatic
    def __type_matches(self, match_type, is_matched):
        if match_type == 'MATCH':
            return is_matched
        elif match_type == 'NOT_MATCH':
            return not is_matched
        else:
            Log.get().debug('Unsupported type[{}]. Please use the latest version of sdk.'.format(match_type))
            return False


class ValueOperatorMatcherFactory(object):

    def __init__(self):
        self._value_matchers = {
            'STRING': StringValueMatcher(),
            'NUMBER': NumberValueMatcher(),
            'BOOLEAN': BoolValueMatcher(),
            'VERSION': VersionValueMatcher(),
            'NULL': NoneValueMatcher(),
            'UNKNOWN': NoneValueMatcher()
        }
        self._operator_matchers = {
            'IN': InMatcher(),
            'CONTAINS': ContainsMatcher(),
            'STARTS_WITH': StartsWithMatcher(),
            'ENDS_WITH': EndsWithMatcher(),
            'GT': GreaterThanMatcher(),
            'GTE': GreaterThanOrEqualToMatcher(),
            'LT': LessThanMatcher(),
            'LTE': LessThanOrEqualToMatcher()
        }

    def get_value_matcher_or_none(self, value_type):
        return self._value_matchers.get(value_type)

    def get_operator_matcher_or_none(self, operator):
        return self._operator_matchers.get(operator)



_C0 = 1.0
_C1 = _C0 * 1000
_C2 = _C1 * 1000
_C3 = _C2 * 1000

_C = [_C0, _C1, _C2, _C3]


class TimeUnit:
    def __init__(self, index):
        """
        :param int index:
        """
        self.__index = index

    @property
    def index(self):
        return self.__index

    def convert(self, source_duration, source_unit):
        """
        :param float source_duration:
        :param TimeUnit source_unit:
        :return:
        """
        return source_duration / (_C[self.index] / _C[source_unit.index])


NANOSECONDS = TimeUnit(0)
MICROSECONDS = TimeUnit(1)
MILLISECONDS = TimeUnit(2)
SECONDS = TimeUnit(3)


def nanos_to_unit(nanos, unit):
    return unit.convert(nanos, NANOSECONDS)



from hackle.internal.evaluation.bucket.bucketer import Bucketer
from hackle.internal.evaluation.evaluator.experiment.experiment_request import ExperimentRequest
from hackle.internal.model.entities import Container


class ContainerResolver(object):

    def __init__(self, bucketer):
        """
        :param Bucketer bucketer:
        """
        self.__bucketer = bucketer

    def is_user_in_container_group(self, request, container):
        """
        :param ExperimentRequest request:
        :param Container container:

        :rtype: bool
        """

        identifier = request.user.identifiers.get(request.experiment.identifier_type)
        if identifier is None:
            return False

        bucket = request.workspace.get_bucket_or_none(container.bucket_id)
        if bucket is None:
            raise Exception('Bucket[{}]'.format(container.bucket_id))

        slot = self.__bucketer.bucketing(bucket, identifier)
        if slot is None:
            return False

        container_group = container.get_group_or_none(slot.variation_id)
        if container_group is None:
            raise Exception('ContainerGroup[{}]'.format(slot.variation_id))

        return request.experiment.id in container_group.experiments



from six.moves.queue import Queue

from . import exceptions as hackle_exceptions
from . import logger as _logging
from . import version
from .commons.enums import Default
from .decision import ExperimentDecision, DecisionReason, FeatureFlagDecision
from .internal.concurrent.schedule.thread_scheduler import ThreadScheduler
from .internal.event.event_dispatcher import EventDispatcher
from .internal.event.event_processor import EventProcessor
from .internal.hackle_core import HackleCore
from .internal.http.http_client import HttpClient
from .internal.logger.log import Log
from .internal.model.sdk import Sdk
from .internal.type import hackle_types
from .internal.user.hackle_user_resolver import HackleUserResolver
from .internal.workspace.workspace_fetcher import WorkspaceFetcher
from .model import HackleEvent, HackleRemoteConfig, PropertyOperations
from .remote_config import HackleRemoteConfigImpl


def __singleton(cls):
    instance = [None]

    def wrapper(*args, **kwargs):
        if instance[0] is None:
            instance[0] = cls(*args, **kwargs)
        return instance[0]

    return wrapper


@__singleton
class Client(object):
    def __init__(self, sdk_key=None, logger=None, timeout=None):
        if sdk_key is None:
            raise hackle_exceptions.RequiredParameterException('sdk_key must not be empty.')

        Log.initialize(_logging.adapt_logger(logger or _logging.NoOpLogger()))

        sdk = Sdk(sdk_key, "python-sdk", version.__version__)

        http_client = HttpClient(sdk)

        workspace_fetcher = WorkspaceFetcher(
            sdk_base_url=Default.SDK_URL,
            http_client=http_client,
            polling_interval_seconds=10,
            scheduler=ThreadScheduler()
        )

        event_processor = EventProcessor(
            queue=Queue(maxsize=1000),
            event_dispatcher=EventDispatcher(Default.EVENT_URL, http_client),
            event_dispatch_size=500,
            flush_scheduler=ThreadScheduler(),
            flush_interval_seconds=10,
            shutdown_timeout_seconds=10
        )

        workspace_fetcher.start()
        event_processor.start()

        self._core = HackleCore.create(workspace_fetcher, event_processor)
        self._user_resolver = HackleUserResolver()

    def close(self):
        self._core.close()

    def __exit__(self):
        self.close()

    def variation(self, experiment_key, user, default_variation='A'):
        """
        Decide the variation to expose to the user for experiment.

        This method return the "A" if:
            - The experiment key is invalid
            - The experiment has not started yet
            - The user is not allocated to the experiment
            - The decided variation has been dropped

        :param int experiment_key: the unique key of the experiment.
        :param hackle.model.User or hackle.model.HackleUser user: the user to participate in the experiment.
        :param str default_variation: the default variation of the experiment.

        :rtype: str
        :return: the decided variation for the user, or the default variation.
        """
        return self.variation_detail(experiment_key, user, default_variation).variation

    def variation_detail(self, experiment_key, user, default_variation='A'):
        """
        Decide the variation to expose to the user for experiment, and returns an object that
        describes the way the variation was decided.

        :param int experiment_key: the unique key of the experiment.
        :param hackle.model.User or hackle.model.HackleUser user: the user to participate in the experiment.
        :param str default_variation: the default variation of the experiment.

        :rtype: hackle.decision.ExperimentDecision
        :return: an object describing the result
        """
        try:
            if not hackle_types.is_positive_int(experiment_key):
                Log.get().warning('Invalid experiment_key: {} (expected: positive integer)'.format(experiment_key))
                return ExperimentDecision(default_variation, DecisionReason.INVALID_INPUT)

            hackle_user = self._user_resolver.resolve_or_none(user)
            if hackle_user is None:
                return ExperimentDecision(default_variation, DecisionReason.INVALID_INPUT)

            return self._core.experiment(experiment_key, hackle_user, default_variation)
        except Exception as e:
            Log.get().error(
                'Unexpected error while deciding variation of experiment[{}]: {}'.format(experiment_key, str(e)))
            return ExperimentDecision(default_variation, DecisionReason.EXCEPTION)

    def is_feature_on(self, feature_key, user):
        """
        Decide whether the feature is turned on to the user.

        :param int feature_key: the unique key for the feature.
        :param hackle.model.User or hackle.model.HackleUser user: the user requesting the feature.

        :rtype: bool
        :return: True if the feature is on
                 False if the feature is off
        """
        return self.feature_flag_detail(feature_key, user).is_on

    def feature_flag_detail(self, feature_key, user):
        """
        Decide whether the feature is turned on to the user, and returns an object that
        describes the way the value was decided.

        :param int feature_key: the unique key of the feature.
        :param hackle.model.User or hackle.model.HackleUser user: the user requesting the feature.

        :rtype: hackle.decision.FeatureFlagDecision
        :return: an object describing the result
        """
        try:
            if not hackle_types.is_positive_int(feature_key):
                Log.get().warning('Invalid feature_key: {} (expected: positive integer)'.format(feature_key))
                return FeatureFlagDecision(False, DecisionReason.INVALID_INPUT)

            hackle_user = self._user_resolver.resolve_or_none(user)
            if hackle_user is None:
                return FeatureFlagDecision(False, DecisionReason.INVALID_INPUT)

            return self._core.feature_flag(feature_key, hackle_user)
        except Exception as e:
            Log.get().error('Unexpected error while deciding feature flag[{}]: {}'.format(feature_key, str(e)))
            return FeatureFlagDecision(False, DecisionReason.EXCEPTION)

    def track(self, event, user):
        """
        Records the event that occurred by the user.

        :param hackle.model.Event or hackle.model.HackleEvent event: the event that occurred.
        :param hackle.model.User or hackle.model.HackleUser user: the user that occurred the event.
        """
        try:
            hackle_event = HackleEvent.from_event(event)
            if not hackle_event.is_valid:
                Log.get().warning('Invalid event for track: [{}]'.format(hackle_event.error_or_none))
                return

            hackle_user = self._user_resolver.resolve_or_none(user)
            if hackle_user is None:
                return

            self._core.track(hackle_event, hackle_user)
        except Exception as e:
            Log.get().error('Unexpected error while tracking event: {}'.format(str(e)))

    def remote_config(self, user):
        """
        Returns a instance of Hackle Remote Config.

        :param hackle.model.User or hackle.model.HackleUser user: the identifier of user.
        :rtype: HackleRemoteConfig
        """
        return HackleRemoteConfigImpl(user, self._core, self._user_resolver)

    def update_user_properties(self, operations, user):
        """
        Updates the user's properties.

        :param hackle.model.PropertyOperations operations: Property operations to update user properties.
        :param hackle.model.HackleUser user: the user whos properties will be updated.
        """
        try:
            if not isinstance(operations, PropertyOperations):
                Log.get().warning(
                    'Invalid user property update operations: {} (expected: PropertyOperations)'.format(operations))
                return

            event = operations.to_event()
            self.track(event, user)
        except Exception as e:
            Log.get().error('Unexpected exception while update user properties: {}'.format(str(e)))



import abc
import time

from six import add_metaclass


@add_metaclass(abc.ABCMeta)
class Clock:

    @abc.abstractmethod
    def current_millis(self):
        """
        :return: epoch millis
        :rtype: int
        """
        pass


class SystemClock(Clock):

    def current_millis(self):
        return int(round(time.time() * 1000))


SYSTEM_CLOCK = SystemClock()



from hackle.internal.evaluation.bucket.bucketer import Bucketer
from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.flow.flow_evaluator import *
from hackle.internal.evaluation.match.condition_matcher_factory import ConditionMatcherFactory
from hackle.internal.evaluation.match.target_matcher import TargetMatcher
from hackle.internal.evaluation.target.experiment_target_determiner import ExperimentTargetDeterminer
from hackle.internal.evaluation.target.experiment_target_rule_determiner import ExperimentTargetRuleDeterminer
from hackle.internal.evaluation.target.override_resolver import OverrideResolver
from hackle.internal.evaluation.target.remote_config_target_rule_determiner import RemoteConfigTargetRuleDeterminer, \
    RemoteConfigTargetRuleMatcher


class EvaluationFlowFactory(object):

    def __init__(self, evaluator):
        """
        :param Evaluator evaluator:
        """
        bucketer = Bucketer()
        target_matcher = TargetMatcher(ConditionMatcherFactory(evaluator))
        action_resolver = ActionResolver(bucketer)
        override_resolver = OverrideResolver(target_matcher, action_resolver)
        container_resolver = ContainerResolver(bucketer)

        self.__ab_test_flow = EvaluationFlow.of(
            OverrideEvaluator(override_resolver),
            IdentifierEvaluator(),
            ContainerEvaluator(container_resolver),
            ExperimentTargetEvaluator(ExperimentTargetDeterminer(target_matcher)),
            DraftEvaluator(),
            PausedEvaluator(),
            CompletedEvaluator(),
            TrafficAllocateEvaluator(action_resolver)
        )

        self.__feature_flag_flow = EvaluationFlow.of(
            DraftEvaluator(),
            PausedEvaluator(),
            CompletedEvaluator(),
            OverrideEvaluator(override_resolver),
            IdentifierEvaluator(),
            TargetRuleEvaluator(ExperimentTargetRuleDeterminer(target_matcher), action_resolver),
            DefaultRuleEvaluator(action_resolver)
        )

        self.__remote_config_parameter_target_rule_determiner = RemoteConfigTargetRuleDeterminer(
            RemoteConfigTargetRuleMatcher(target_matcher, bucketer))

    def get_evaluation_flow(self, experiment_type):
        if experiment_type == 'AB_TEST':
            return self.__ab_test_flow
        elif experiment_type == 'FEATURE_FLAG':
            return self.__feature_flag_flow
        else:
            raise Exception('Unsupported type [{}]'.format(experiment_type))

    @property
    def remote_config_parameter_target_rule_determiner(self):
        """
        :rtype: RemoteConfigTargetRuleDeterminer 
        """
        return self.__remote_config_parameter_target_rule_determiner



import os
from unittest import TestCase

from hackle.internal.workspace.workspace import Workspace


class WorkspaceTest(TestCase):

    def test_workspace_fetch(self):
        with open(os.path.join(os.path.dirname(__file__), '../../resources/workspace_config.json'), 'r') as f:
            data = f.read()
        workspace = Workspace(data)
        self.assertIsNotNone(workspace)
        self.assertTrue(bool(workspace.bucket_id_map))
        self.assertTrue(bool(workspace.event_type_key_map))
        self.assertTrue(bool(workspace.experiment_key_map))
        self.assertTrue(bool(workspace.feature_flag_key_map))
        self.assertTrue(bool(workspace.parameter_configurations))
        self.assertTrue(bool(workspace.remote_config_parameters))
        self.assertTrue(bool(workspace.segment_key_map))



from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.evaluator.experiment.experiment_request import ExperimentRequest
from hackle.internal.evaluation.match.target_matcher import TargetMatcher


class ExperimentTargetDeterminer(object):
    def __init__(self, target_matcher):
        """
        :param TargetMatcher target_matcher:
        """
        self.__target_matcher = target_matcher

    def is_user_in_experiment_target(self, request, context):
        """
        :param ExperimentRequest request:
        :param Evaluator.Context context:
        :rtype: bool
        """
        if not request.experiment.target_audiences:
            return True

        for target in request.experiment.target_audiences:
            if self.__target_matcher.matches(request, context, target):
                return True

        return False



from unittest import TestCase
from unittest.mock import Mock

from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.match.segment_condition_matcher import SegmentMatcher, SegmentConditionMatcher
from hackle.internal.model.entities import *
from tests.internal.evaluation.evaluator.experiment.test_experiment_request import create_experiment_request


class SegmentConditionMatcherTest(TestCase):

    def setUp(self):
        self.segment_matcher = Mock(spec=SegmentMatcher)
        self.sut = SegmentConditionMatcher(self.segment_matcher)

    def test_keyType이_SEGMENT_타입이_아니면_예외가_발생한다(self):
        # given
        condition = TargetCondition(
            TargetKey('USER_PROPERTY', "age"),
            TargetMatch('MATCH', 'IN', 'NUMBER', [30])
        )

        # when
        with self.assertRaises(Exception) as actual:
            self.sut.matches(Mock(), Mock(), condition)

        # then
        self.assertEqual('Unsupported target.key.type [USER_PROPERTY]', str(actual.exception))

    def test_등록된_segmentKey가_String_타입이_아니면_예외가_발생한다(self):
        # given
        condition = TargetCondition(
            TargetKey('SEGMENT', "SEGMENT"),
            TargetMatch('MATCH', 'IN', 'NUMBER', [30])
        )

        # when
        with self.assertRaises(Exception) as actual:
            self.sut.matches(Mock(), Mock(), condition)

        # then
        self.assertEqual('SegmentKey[30]', str(actual.exception))

    def test_등록된_segmentKey에_해당하는_Segment가_없으면_예외가_발생한다(self):
        # given
        condition = TargetCondition(
            TargetKey('SEGMENT', "SEGMENT"),
            TargetMatch('MATCH', 'IN', 'STRING', ["seg_01"])
        )
        workspace = WorkspaceStub()
        request = create_experiment_request(workspace=workspace)

        # when
        with self.assertRaises(Exception) as actual:
            self.sut.matches(request, Evaluator.context(), condition)

        # then
        self.assertEqual('Segment[seg_01]', str(actual.exception))

    def test_등록된_segment_중_일치하는게_하나라도_있으면_true(self):
        # given
        condition = TargetCondition(
            TargetKey('SEGMENT', "SEGMENT"),
            TargetMatch('MATCH', 'IN', 'STRING', ["seg_01"])
        )
        workspace = WorkspaceStub()
        workspace.add_segment("seg_01")
        self.segment_matcher.matches.return_value = True

        request = create_experiment_request(workspace=workspace)

        # when
        actual = self.sut.matches(request, Evaluator.context(), condition)

        # then
        self.assertTrue(actual)

    def test_TC_MatchType(self):
        # given
        condition = TargetCondition(
            TargetKey('SEGMENT', "SEGMENT"),
            TargetMatch('NOT_MATCH', 'IN', 'STRING', ["seg_01"])
        )
        workspace = WorkspaceStub()
        workspace.add_segment("seg_01")
        self.segment_matcher.matches.return_value = True

        request = create_experiment_request(workspace=workspace)

        # when
        actual = self.sut.matches(request, Evaluator.context(), condition)

        # then
        self.assertFalse(actual)

    def test_등록된_segment_중_일치하는게_없으면_false(self):
        # given
        condition = TargetCondition(
            TargetKey('SEGMENT', "SEGMENT"),
            TargetMatch('MATCH', 'IN', 'STRING', ["seg_01", "seg_02"])
        )
        workspace = WorkspaceStub()
        workspace.add_segment("seg_01")
        workspace.add_segment("seg_02")
        self.segment_matcher.matches.return_value = False

        request = create_experiment_request(workspace=workspace)

        # when
        actual = self.sut.matches(request, Evaluator.context(), condition)

        # then
        self.assertFalse(actual)


class WorkspaceStub:

    def __init__(self):
        self.segments = {}

    def add_segment(self, segment_key):
        self.segments[segment_key] = Mock()

    def get_segment_or_none(self, segment_key):
        return self.segments.get(segment_key)



from unittest import TestCase, mock
from unittest.mock import Mock

from hackle.internal.http.http_client import HttpClient
from hackle.internal.model.sdk import Sdk
from tests.internal.time.test_clock import FixedClock


class HttpClientTest(TestCase):

    @mock.patch('requests.get')
    def test_get(self, mock_get):
        sdk = Sdk('test_key', 'test_name', 'test_version')
        clock = FixedClock(42)
        sut = HttpClient(sdk, clock)

        response = Mock()
        response.status_code = 200
        response.content = bytes("{'a': 'b'}", 'utf-8')

        mock_get.return_value = response

        actual = sut.get('localhost')

        args = mock_get.call_args
        self.assertEqual('localhost', args[0][0])
        self.assertEqual(
            {
                'X-HACKLE-SDK-KEY': 'test_key',
                'X-HACKLE-SDK-NAME': 'test_name',
                'X-HACKLE-SDK-VERSION': 'test_version',
                'X-HACKLE-SDK-TIME': '42'
            },
            args[1]['headers']
        )
        self.assertEqual("{'a': 'b'}", actual)

    @mock.patch('requests.post')
    def test_post(self, mock_post):
        sdk = Sdk('test_key', 'test_name', 'test_version')
        clock = FixedClock(42)
        sut = HttpClient(sdk, clock)

        response = Mock()
        response.status_code = 200

        sut.post('localhost', "{'a': 'b'}")

        args = mock_post.call_args
        self.assertEqual('localhost', args[0][0])

        self.assertEqual("{'a': 'b'}", args[1]['data'])
        self.assertEqual(
            {
                'X-HACKLE-SDK-KEY': 'test_key',
                'X-HACKLE-SDK-NAME': 'test_name',
                'X-HACKLE-SDK-VERSION': 'test_version',
                'X-HACKLE-SDK-TIME': '42',
                'Content-Type': 'application/json'
            },
            args[1]['headers']
        )



import requests

from hackle.internal.model.sdk import Sdk
from hackle.internal.time.clock import SYSTEM_CLOCK, Clock


class HttpClient(object):
    __SDK_KEY_HEADER_NAME = 'X-HACKLE-SDK-KEY'
    __SDK_NAME_HEADER_NAME = 'X-HACKLE-SDK-NAME'
    __SDK_VERSION_HEADER_NAME = 'X-HACKLE-SDK-VERSION'
    __SDK_TIME_HEADER_NAME = 'X-HACKLE-SDK-TIME'
    __CONTENT_TYPE_HEADER_NAME = 'Content-Type'
    __CONTENT_TYPE_VALUE = 'application/json'
    __TIMEOUT_SECONDS = 10

    def __init__(self, sdk, clock=SYSTEM_CLOCK):
        """
        :param Sdk sdk:
        :param Clock clock:
        """
        self.__sdk = sdk
        self.__clock = clock

    def get(self, url):
        response = requests.get(url, headers=self.__headers(), timeout=self.__TIMEOUT_SECONDS)
        response.raise_for_status()
        return response.content.decode('utf-8')

    def post(self, url, data):
        headers = self.__headers()
        headers[self.__CONTENT_TYPE_HEADER_NAME] = self.__CONTENT_TYPE_VALUE
        requests.post(url, data=data, headers=headers, timeout=self.__TIMEOUT_SECONDS).raise_for_status()

    def __headers(self):
        return {
            self.__SDK_KEY_HEADER_NAME: self.__sdk.key,
            self.__SDK_NAME_HEADER_NAME: self.__sdk.name,
            self.__SDK_VERSION_HEADER_NAME: self.__sdk.version,
            self.__SDK_TIME_HEADER_NAME: str(self.__clock.current_millis())
        }



from unittest import TestCase
from unittest.mock import Mock

from hackle.decision import DecisionReason
from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.evaluator.experiment.experiment_evaluation import ExperimentEvaluation
from hackle.internal.evaluation.flow.evaluation_flow import EvaluationFlow
from hackle.internal.evaluation.flow.flow_evaluator import FlowEvaluator
from tests.internal.evaluation.evaluator.experiment.test_experiment_request import create_experiment_request


class EvaluationFlowTest(TestCase):

    def test__evaluate__end(self):
        request = create_experiment_request()
        evaluation = EvaluationFlow().evaluate(request, Evaluator.context())
        self.assertEqual(DecisionReason.TRAFFIC_NOT_ALLOCATED, evaluation.reason)
        self.assertEqual('A', evaluation.variation_key)

    def test__evaluate__flow(self):
        evaluation = Mock(spec=ExperimentEvaluation)
        next_flow = Mock(spec=EvaluationFlow)
        flow_evaluator = Mock(spec=FlowEvaluator)
        flow_evaluator.evaluate.return_value = evaluation

        request = create_experiment_request()

        sut = EvaluationFlow(flow_evaluator, next_flow)

        actual = sut.evaluate(request, Evaluator.context())

        self.assertEqual(evaluation, actual)
        flow_evaluator.evaluate.assert_called_once()

    def test_of(self):
        fe1 = Mock(spec=FlowEvaluator)
        fe2 = Mock(spec=FlowEvaluator)
        fe3 = Mock(spec=FlowEvaluator)

        flow = EvaluationFlow.of(fe1, fe2, fe3)

        self.assertFalse(flow.is_end)
        self.assertEqual(fe1, flow.flow_evaluator)

        flow = flow.next_flow
        self.assertFalse(flow.is_end)
        self.assertEqual(fe2, flow.flow_evaluator)

        flow = flow.next_flow
        self.assertFalse(flow.is_end)
        self.assertEqual(fe3, flow.flow_evaluator)

        flow = flow.next_flow
        self.assertTrue(flow.is_end)



from hackle.internal.user.hackle_user_resolver import HackleUserResolver
from hackle.internal.user.internal_hackle_user import InternalHackleUser
from hackle.model import User, HackleUser
from tests import base


class HackleUserResolverTest(base.BaseTest):

    def setUp(self):
        self.sut = HackleUserResolver()

    def test_user가_None이면_None리턴(self):
        # given
        user = None

        # when
        actual = self.sut.resolve_or_none(user)

        # then
        self.assertIsNone(actual)

    def test_user가_User_or_HackleUser_타입이_아니면_None_리턴(self):
        # given
        user = 'string user'

        # when
        actual = self.sut.resolve_or_none(user)

        # then
        self.assertIsNone(actual)

    def test_user가_User_타입이면_HackleUser로_변환하여_처리한다(self):
        # given
        user = User(id='test_id', properties={'age': 30, 'grade': 'GOLD'})

        # when
        actual = self.sut.resolve_or_none(user)

        # then
        self.assertEqual(
            InternalHackleUser(identifiers={'$id': 'test_id'}, properties={'age': 30, 'grade': 'GOLD'}),
            actual
        )

    def test_user가_HackleUser_타입이면_그대로_처리한다(self):
        # given
        user = HackleUser(
            id='test_id',
            user_id='test_user_id',
            device_id='test_device_id',
            identifiers={
                'custom_id': 'test_custom_id'
            },
            properties={
                'grade': 'GOLD',
                'age': 30
            }
        )

        # when
        actual = self.sut.resolve_or_none(user)

        # then
        self.assertEqual(
            InternalHackleUser(
                identifiers={
                    '$id': 'test_id',
                    '$userId': 'test_user_id',
                    '$deviceId': 'test_device_id',
                    'custom_id': 'test_custom_id'
                },
                properties={
                    'age': 30,
                    'grade': 'GOLD'
                }
            ),
            actual
        )

    def test_유효하지않은_properties는_필터링한다(self):
        # given
        user = HackleUser(
            id='test_id',
            user_id='test_user_id',
            device_id='test_device_id',
            identifiers={
                'custom_id': 'test_custom_id'
            },
            properties={
                'grade': 'GOLD',
                'age': 30,
                'a' * 129: 'too long key',
                'object_value_is_invalid': {1: 1}
            }
        )

        # when
        actual = self.sut.resolve_or_none(user)

        # then
        self.assertEqual(
            InternalHackleUser(
                identifiers={
                    '$id': 'test_id',
                    '$userId': 'test_user_id',
                    '$deviceId': 'test_device_id',
                    'custom_id': 'test_custom_id'
                },
                properties={
                    'age': 30,
                    'grade': 'GOLD'
                }
            ),
            actual
        )

    def test_id가_없는_경우(self):
        # given
        user = HackleUser(
            user_id='test_user_id',
            device_id='test_device_id',
            identifiers={
                'custom_id': 'test_custom_id'
            },
            properties={
                'grade': 'GOLD',
                'age': 30
            }
        )

        # when
        actual = self.sut.resolve_or_none(user)

        # then
        self.assertEqual(
            InternalHackleUser(
                identifiers={
                    '$userId': 'test_user_id',
                    '$deviceId': 'test_device_id',
                    'custom_id': 'test_custom_id'
                },
                properties={
                    'age': 30,
                    'grade': 'GOLD'
                }
            ),
            actual
        )

    def test_user_id가_없는_경우(self):
        # given
        user = HackleUser(
            id='test_id',
            device_id='test_device_id',
            identifiers={
                'custom_id': 'test_custom_id'
            },
            properties={
                'grade': 'GOLD',
                'age': 30
            }
        )

        # when
        actual = self.sut.resolve_or_none(user)

        # then
        self.assertEqual(
            InternalHackleUser(
                identifiers={
                    '$id': 'test_id',
                    '$deviceId': 'test_device_id',
                    'custom_id': 'test_custom_id'
                },
                properties={
                    'age': 30,
                    'grade': 'GOLD'
                }
            ),
            actual
        )

    def test_device_id가_없는_경우(self):
        # given
        user = HackleUser(
            id='test_id',
            user_id='test_user_id',
            identifiers={
                'custom_id': 'test_custom_id'
            },
            properties={
                'grade': 'GOLD',
                'age': 30
            }
        )

        # when
        actual = self.sut.resolve_or_none(user)

        # then
        self.assertEqual(
            InternalHackleUser(
                identifiers={
                    '$id': 'test_id',
                    '$userId': 'test_user_id',
                    'custom_id': 'test_custom_id'
                },
                properties={
                    'age': 30,
                    'grade': 'GOLD'
                }
            ),
            actual
        )

    def test_식별자가_하나도_없는경우_None_리턴(self):
        # given
        user = HackleUser(
            properties={
                'grade': 'GOLD',
                'age': 30
            }
        )

        # when
        actual = self.sut.resolve_or_none(user)

        # then
        self.assertIsNone(actual)

    def test_propertis가_None인_경우(self):
        # given
        user = HackleUser(
            id='test_id',
            user_id='test_user_id',
            identifiers={
                'custom_id': 'test_custom_id'
            }
        )

        # when
        actual = self.sut.resolve_or_none(user)

        # then
        self.assertEqual(
            InternalHackleUser(
                identifiers={
                    '$id': 'test_id',
                    '$userId': 'test_user_id',
                    'custom_id': 'test_custom_id'
                },
                properties={}
            ),
            actual
        )



from unittest import TestCase
from unittest.mock import Mock

from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.match.experiment_condition_matcher import AbTestConditionMatcher, \
    FeatureFlagConditionMatcher, \
    ExperimentConditionMatcher
from hackle.internal.model.entities import TargetCondition, TargetKey, TargetMatch
from tests.internal.evaluation.evaluator.experiment.test_experiment_request import create_experiment_request
from tests.internal.model.test_entities import create_experiment


class ExperimentConditionMatcherTest(TestCase):

    def setUp(self):
        self.ab_test_matcher = Mock(spec=AbTestConditionMatcher)
        self.feature_flag_matcher = Mock(spec=FeatureFlagConditionMatcher)
        self.sut = ExperimentConditionMatcher(self.ab_test_matcher, self.feature_flag_matcher)

    def test_ab_test(self):
        # given
        request = create_experiment_request(experiment=create_experiment(type='AB_TEST'))
        condition = TargetCondition(
            TargetKey('AB_TEST', '42'),
            TargetMatch('MATCH', 'IN', 'STRING', ['A'])
        )

        self.ab_test_matcher.matches.return_value = True

        # when
        actual = self.sut.matches(request, Evaluator.context(), condition)

        # then
        self.assertTrue(actual)
        self.ab_test_matcher.matches.assert_called_once()

    def test_feature_flag(self):
        # given
        request = create_experiment_request(experiment=create_experiment(type='FEATURE_FLAG'))
        condition = TargetCondition(
            TargetKey('FEATURE_FLAG', '42'),
            TargetMatch('MATCH', 'IN', 'BOOLEAN', [True])
        )

        self.ab_test_matcher.matches.return_value = True

        # when
        actual = self.sut.matches(request, Evaluator.context(), condition)

        # then
        self.assertTrue(actual)
        self.feature_flag_matcher.matches.assert_called_once()

    def test_unsupported(self):
        request = create_experiment_request(experiment=create_experiment())

        def verify(key_type):
            condition = TargetCondition(
                TargetKey(key_type, '42'),
                TargetMatch('MATCH', 'IN', 'STRING', ['A'])
            )
            with self.assertRaises(Exception):
                self.sut.matches(request, Evaluator.context(), condition)

        verify('USER_ID')
        verify('USER_PROPERTY')
        verify('HACKLE_PROPERTY')
        verify('SEGMENT')



from hackle.decision import DecisionReason
from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.evaluator.experiment.experiment_evaluation import ExperimentEvaluation
from hackle.internal.evaluation.evaluator.experiment.experiment_request import ExperimentRequest


class EvaluationFlow(object):

    def __init__(self, flow_evaluator=None, next_flow=None):
        """
        :param hackle.internal.evaluation.flow.flow_evaluator.FlowEvaluator or None flow_evaluator:
        :param EvaluationFlow or None next_flow:
        """
        self.__flow_evaluator = flow_evaluator
        self.__next_flow = next_flow

    @staticmethod
    def of(*flow_evaluators):
        flow = EvaluationFlow()
        for flow_evaluator in reversed(flow_evaluators):
            flow = EvaluationFlow(flow_evaluator, flow)
        return flow

    @property
    def flow_evaluator(self):
        """
        :rtype: hackle.internal.evaluation.flow.flow_evaluator.FlowEvaluator or None
        """
        return self.__flow_evaluator

    @property
    def next_flow(self):
        """
        :rtype: EvaluationFlow or None
        """
        return self.__next_flow

    @property
    def is_end(self):
        """
        :rtype: bool
        """
        return self.__flow_evaluator is None or self.__next_flow is None

    def evaluate(self, request, context):
        """
        :param ExperimentRequest request:
        :param Evaluator.Context context:

        :rtype: ExperimentEvaluation
        """
        if self.is_end:
            return ExperimentEvaluation.of_default(request, context, DecisionReason.TRAFFIC_NOT_ALLOCATED)
        else:
            return self.__flow_evaluator.evaluate(request, context, self.__next_flow)



from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.evaluator.experiment.experiment_request import ExperimentRequest
from hackle.internal.evaluation.match.target_matcher import TargetMatcher
from hackle.internal.model.entities import TargetRule


class ExperimentTargetRuleDeterminer(object):
    def __init__(self, target_matcher):
        """
        :param TargetMatcher target_matcher: 
        """
        self.__target_matcher = target_matcher

    def determine_target_rule_or_none(self, request, context):
        """
        :param ExperimentRequest request: 
        :param Evaluator.Context context: 
        :rtype: TargetRule 
        """
        for target_rule in request.experiment.target_rules:
            if self.__target_matcher.matches(request, context, target_rule.target):
                return target_rule

        return None



from hackle.internal.workspace.workspace import Workspace
from tests.resources.test_resource_reader import read_resource


class ResourceWorkspaceFetcher(object):

    def __init__(self, file_name):
        data = read_resource(file_name)
        self.__workspace = Workspace(data)

    def fetch(self):
        return self.__workspace



class Sdk(object):
    def __init__(self, key, name, version):
        """
        :param str key:
        :param str name:
        :param str version:
        """
        self.__key = key
        self.__name = name
        self.__version = version

    @property
    def key(self):
        """
        :rtype: str
        """
        return self.__key

    @property
    def name(self):
        """
        :rtype: str
        """
        return self.__name

    @property
    def version(self):
        """
        :rtype: str
        """
        return self.__version




from unittest import TestCase
from unittest.mock import Mock

from hackle.internal.evaluation.evaluator.delegating_evaluator import DelegatingEvaluator
from hackle.internal.evaluation.evaluator.evaluator import Evaluator, ContextualEvaluator


class DelegatingEvaluatorTest(TestCase):

    def test_evaluate(self):
        sut = DelegatingEvaluator()

        r1 = Mock(spec=Evaluator.Request)
        e1 = Mock(spec=Evaluator.Evaluation)
        evaluator1 = MockEvaluator(r1, e1)

        sut.add(evaluator1)
        with self.assertRaises(Exception):
            sut.evaluate(Mock(), Mock())
        self.assertEqual(e1, sut.evaluate(r1, Evaluator.context()))

        r2 = Mock(spec=Evaluator.Request)
        e2 = Mock(spec=Evaluator.Evaluation)
        evaluator2 = MockEvaluator(r2, e2)

        sut.add(evaluator2)
        self.assertEqual(e2, sut.evaluate(r2, Evaluator.context()))


class MockEvaluator(ContextualEvaluator):

    def __init__(self, request, evaluation):
        super(MockEvaluator, self).__init__()
        self.__request = request
        self.__evaluation = evaluation

    def supports(self, request):
        return request == self.__request

    def evaluate_internal(self, request, context):
        return self.__evaluation



import abc
import uuid

from six import add_metaclass

from hackle.internal.evaluation.evaluator.experiment.experiment_evaluation import ExperimentEvaluation
from hackle.internal.evaluation.evaluator.remoteconfig.remote_config_evaluation import RemoteConfigEvaluation
from hackle.internal.model.entities import EventType, Experiment, RemoteConfigParameter
from hackle.internal.user.internal_hackle_user import InternalHackleUser
from hackle.model import HackleEvent


@add_metaclass(abc.ABCMeta)
class UserEvent(object):

    def __init__(self, insert_id, timestamp, user):
        """
        :param str insert_id:
        :param int timestamp:
        :param InternalHackleUser user:
        """
        self.__insert_id = insert_id
        self.__timestamp = timestamp
        self.__user = user

    @property
    def insert_id(self):
        """
        :rtype: str
        """
        return self.__insert_id

    @property
    def timestamp(self):
        """
        :rtype: int
        """
        return self.__timestamp

    @property
    def user(self):
        """
        :rtype: InternalHackleUser
        """
        return self.__user

    @staticmethod
    def exposure(user, evaluation, properties, timestamp):
        """
        :param InternalHackleUser user:
        :param ExperimentEvaluation evaluation:
        :param dict[str, object] properties:
        :param int timestamp:
        :rtype: ExposureEvent
        """
        return ExposureEvent(
            UserEvent.__insert_id(),
            timestamp,
            user,
            evaluation.experiment,
            evaluation.variation_id,
            evaluation.variation_key,
            evaluation.reason,
            properties
        )

    @staticmethod
    def track(user, event_type, event, timestamp):
        """
        :param InternalHackleUser user:
        :param EventType event_type:
        :param HackleEvent event:
        :param int timestamp:

        :rtype: TrackEvent
        """
        return TrackEvent(
            UserEvent.__insert_id(),
            timestamp,
            user,
            event_type,
            event
        )

    @staticmethod
    def remote_config(user, evaluation, properties, timestamp):
        """
        :param InternalHackleUser user:
        :param RemoteConfigEvaluation evaluation:
        :param dict[str, object] properties:
        :param int timestamp:
        :rtype: RemoteConfigEvent
        """
        return RemoteConfigEvent(
            UserEvent.__insert_id(),
            timestamp,
            user,
            evaluation.parameter,
            evaluation.value_id,
            evaluation.reason,
            properties
        )

    @staticmethod
    def __insert_id():
        """
        :rtype: str
        """
        return str(uuid.uuid4())


class ExposureEvent(UserEvent):

    def __init__(self, insert_id, timestamp, user, experiment, variation_id, variation_key, reason, properties):
        """
        :param str insert_id:
        :param int timestamp:
        :param InternalHackleUser user:
        :param Experiment experiment:
        :param int or None variation_id:
        :param str variation_key:
        :param str reason:
        :param dict[str, object] properties:
        """
        super(ExposureEvent, self).__init__(insert_id, timestamp, user)
        self.experiment = experiment
        self.variation_id = variation_id
        self.variation_key = variation_key
        self.reason = reason
        self.properties = properties


class TrackEvent(UserEvent):

    def __init__(self, insert_id, timestamp, user, event_type, event):
        """
        :param str insert_id:
        :param int timestamp:
        :param InternalHackleUser user:
        :param EventType event_type:
        :param HackleEvent event:
        """
        super(TrackEvent, self).__init__(insert_id, timestamp, user)
        self.event_type = event_type
        self.event = event


class RemoteConfigEvent(UserEvent):

    def __init__(self, insert_id, timestamp, user, parameter, value_id, reason, properties):
        """
        :param str insert_id:
        :param int timestamp:
        :param InternalHackleUser user:
        :param RemoteConfigParameter parameter:
        :param int or None value_id:
        :param str reason:
        :param dict[str, object] properties:
        """
        super(RemoteConfigEvent, self).__init__(insert_id, timestamp, user)
        self.parameter = parameter
        self.value_id = value_id
        self.reason = reason
        self.properties = properties



from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.model.entities import Experiment
from hackle.internal.user.internal_hackle_user import InternalHackleUser
from hackle.internal.workspace.workspace import Workspace


class ExperimentRequest(Evaluator.Request):
    def __init__(self, workspace, user, experiment, default_variation_key):
        """
        :param Workspace workspace:
        :param InternalHackleUser user:
        :param Experiment experiment:
        :param str default_variation_key:
        """
        super(ExperimentRequest, self).__init__(Evaluator.Key('EXPERIMENT', experiment.id), workspace, user)
        self.__experiment = experiment
        self.__default_variation_key = default_variation_key

    @staticmethod
    def of(workspace, user, experiment, default_variation_key):
        """
        :param Workspace workspace:
        :param InternalHackleUser user:
        :param Experiment experiment:
        :param str default_variation_key:

        :rtype: ExperimentRequest
        """
        return ExperimentRequest(workspace, user, experiment, default_variation_key)

    @staticmethod
    def from_request(requested_by, experiment):
        """
        :param Evaluator.Request requested_by:
        :param Experiment experiment:

        :rtype: ExperimentRequest
        """
        return ExperimentRequest(requested_by.workspace, requested_by.user, experiment, 'A')

    @property
    def experiment(self):
        """
        :rtype: Experiment
        """
        return self.__experiment

    @property
    def default_variation_key(self):
        """
        :rtype: str
        """
        return self.__default_variation_key

    def __str__(self):
        return 'EvaluatorRequest(type={}, key={})'.format(self.__experiment.type, self.__experiment.key)

    def __repr__(self):
        return self.__str__()



from unittest import TestCase
from unittest.mock import Mock

from hackle.decision import ExperimentDecision, FeatureFlagDecision, DecisionReason
from hackle.internal.evaluation.evaluator.experiment.experiment_evaluation import ExperimentEvaluation
from hackle.internal.evaluation.evaluator.experiment.experiment_evaluator import ExperimentEvaluator
from hackle.internal.evaluation.evaluator.remoteconfig.remote_config_evaluation import RemoteConfigEvaluation
from hackle.internal.evaluation.evaluator.remoteconfig.remote_config_evaluator import RemoteConfigEvaluator
from hackle.internal.event.event_processor import EventProcessor
from hackle.internal.event.user_event_factory import UserEventFactory
from hackle.internal.hackle_core import HackleCore
from hackle.internal.model.entities import EventType, ParameterConfiguration
from hackle.internal.time.clock import SYSTEM_CLOCK
from hackle.internal.user.identifier_type import IdentifierType
from hackle.internal.user.internal_hackle_user import InternalHackleUser
from hackle.internal.workspace.workspace import Workspace
from hackle.internal.workspace.workspace_fetcher import WorkspaceFetcher
from hackle.model import HackleEvent
from tests.internal.model.test_entities import create_experiment, create_parameter


class HackleCoreTest(TestCase):

    def setUp(self):
        self.experiment_evaluator = Mock(spec=ExperimentEvaluator)
        self.remote_config_evaluator = Mock(spec=RemoteConfigEvaluator)
        self.workspace_fetcher = Mock(spec=WorkspaceFetcher)
        self.event_factory = Mock(spec=UserEventFactory)
        self.event_processor = Mock(spec=EventProcessor)
        self.clock = SYSTEM_CLOCK
        self.sut = HackleCore(
            self.experiment_evaluator,
            self.remote_config_evaluator,
            self.workspace_fetcher,
            self.event_factory,
            self.event_processor
        )

        self.user = InternalHackleUser.builder().identifier(IdentifierType.ID, "user").build()

    def test_close(self):
        self.sut.close()
        self.workspace_fetcher.stop.assert_called_once()
        self.event_processor.stop.assert_called_once()

    def test__experiment__설정정보를_가져오지_못한_경우_기본그룹으로_결정하고_노출이벤트는_전송하지_않는다(self):
        # given
        self.workspace_fetcher.fetch.return_value = None

        # when
        actual = self.sut.experiment(42, self.user, 'D')

        # then
        self.assertEqual(ExperimentDecision('D', DecisionReason.SDK_NOK_READY), actual)
        self.event_processor.process.assert_not_called()

    def test__experiment__실험키에_해당하는_실험이_없는경우_기본그룹으로_결정하고_노출이벤트는_전송하지_않는다(self):
        # given
        workspace = Workspace('{}')
        self.workspace_fetcher.fetch.return_value = workspace

        # when
        actual = self.sut.experiment(42, self.user, 'G')

        # then
        self.assertEqual(ExperimentDecision('G', DecisionReason.EXPERIMENT_NOT_FOUND), actual)
        self.event_processor.process.assert_not_called()

    def test__experiment__평가결과로_노출이벤트를_전송하고_평가된_그룹으로_결정한다(self):
        # given
        experiment = create_experiment()

        workspace = Mock(spec=Workspace)
        workspace.get_experiment_or_none.return_value = experiment
        self.workspace_fetcher.fetch.return_value = workspace

        config = Mock(spec=ParameterConfiguration)
        config.id = 420

        evaluation = ExperimentEvaluation(
            DecisionReason.TRAFFIC_ALLOCATED,
            [],
            experiment,
            320,
            'B',
            config
        )

        self.experiment_evaluator.evaluate.return_value = evaluation
        self.event_factory.create.return_value = [Mock(), Mock()]

        # when
        actual = self.sut.experiment(42, self.user, 'J')

        # then
        self.assertEqual('B', actual.variation)
        self.assertEqual(DecisionReason.TRAFFIC_ALLOCATED, actual.reason)
        self.assertEqual(config, actual.config)
        self.assertEqual(2, self.event_processor.process.call_count)

    def test__experiment__평가결과로_Config_가_없으면_Empty_config(self):
        # given
        experiment = create_experiment()

        workspace = Mock(spec=Workspace)
        workspace.get_experiment_or_none.return_value = experiment
        self.workspace_fetcher.fetch.return_value = workspace

        evaluation = ExperimentEvaluation(
            DecisionReason.TRAFFIC_ALLOCATED,
            [],
            experiment,
            320,
            'B',
            None
        )

        self.experiment_evaluator.evaluate.return_value = evaluation
        self.event_factory.create.return_value = [Mock(), Mock()]

        # when
        actual = self.sut.experiment(42, self.user, 'J')

        # then
        self.assertEqual('B', actual.variation)
        self.assertEqual(DecisionReason.TRAFFIC_ALLOCATED, actual.reason)
        self.assertEqual({}, actual.config)

    def test__feature_flag__설정정보를_가져오지_못하면_OFF로_결정하고_노출이벤트는_전송하지_않는다(self):
        # given
        self.workspace_fetcher.fetch.return_value = None

        # when
        actual = self.sut.feature_flag(42, self.user)

        # then
        self.assertEqual(FeatureFlagDecision(False, DecisionReason.SDK_NOK_READY), actual)
        self.event_processor.process.assert_not_called()

    def test__feature_flag__기능키에_해당하는_기능플래그가_없는경우_OFF로_결정하고_노출이벤트는_전송하지_않는다(self):
        # given
        workspace = Mock(spec=Workspace)
        workspace.get_feature_flag_or_none.return_value = None

        self.workspace_fetcher.fetch.return_value = workspace

        # when
        actual = self.sut.feature_flag(42, self.user)

        # then
        self.assertEqual(FeatureFlagDecision(False, DecisionReason.FEATURE_FLAG_NOT_FOUND), actual)
        self.event_processor.process.assert_not_called()

    def test__feature_flag__평가결과로_노출이벤트를_전송한다(self):
        # given
        feature_flag = create_experiment(type='FEATURE_FLAG')

        workspace = Mock(spec=Workspace)
        workspace.get_feature_flag_or_none.return_value = feature_flag
        self.workspace_fetcher.fetch.return_value = workspace

        config = Mock(spec=ParameterConfiguration)
        config.id = 420

        evaluation = ExperimentEvaluation(
            DecisionReason.TRAFFIC_ALLOCATED,
            [],
            feature_flag,
            320,
            'A',
            config
        )

        self.experiment_evaluator.evaluate.return_value = evaluation
        self.event_factory.create.return_value = [Mock(), Mock()]

        # when
        self.sut.feature_flag(42, self.user)

        # then
        self.assertEqual(2, self.event_processor.process.call_count)

    def test__feature_flag__그룹_A로_평가되면_OFF로_결정한다(self):
        # given
        feature_flag = create_experiment(type='FEATURE_FLAG')

        workspace = Mock(spec=Workspace)
        workspace.get_feature_flag_or_none.return_value = feature_flag
        self.workspace_fetcher.fetch.return_value = workspace

        evaluation = ExperimentEvaluation(
            DecisionReason.DEFAULT_RULE,
            [],
            feature_flag,
            320,
            'A',
            None
        )

        self.experiment_evaluator.evaluate.return_value = evaluation
        self.event_factory.create.return_value = [Mock(), Mock()]

        # when
        actual = self.sut.feature_flag(42, self.user)

        # then
        self.assertEqual(False, actual.is_on)
        self.assertEqual(DecisionReason.DEFAULT_RULE, actual.reason)

    def test__feature_flag__그룹_A로_평가되지_않으면_ON으로_결정한다(self):
        # given
        feature_flag = create_experiment(type='FEATURE_FLAG')

        workspace = Mock(spec=Workspace)
        workspace.get_feature_flag_or_none.return_value = feature_flag
        self.workspace_fetcher.fetch.return_value = workspace

        evaluation = ExperimentEvaluation(
            DecisionReason.TARGET_RULE_MATCH,
            [],
            feature_flag,
            320,
            'B',
            None
        )

        self.experiment_evaluator.evaluate.return_value = evaluation
        self.event_factory.create.return_value = [Mock(), Mock()]

        # when
        actual = self.sut.feature_flag(42, self.user)

        # then
        self.assertEqual(True, actual.is_on)
        self.assertEqual(DecisionReason.TARGET_RULE_MATCH, actual.reason)

    def test__feature_flag__Config(self):
        # given
        feature_flag = create_experiment(type='FEATURE_FLAG')

        workspace = Mock(spec=Workspace)
        workspace.get_feature_flag_or_none.return_value = feature_flag
        self.workspace_fetcher.fetch.return_value = workspace

        config = Mock(spec=ParameterConfiguration)
        config.id = 420

        evaluation = ExperimentEvaluation(
            DecisionReason.TRAFFIC_ALLOCATED,
            [],
            feature_flag,
            320,
            'A',
            config
        )
        self.experiment_evaluator.evaluate.return_value = evaluation
        self.event_factory.create.return_value = [Mock(), Mock()]

        # when
        actual = self.sut.feature_flag(42, self.user)

        # then
        self.assertEqual(config, actual.config)

    def test__feature_flag__Config_None(self):
        # given
        feature_flag = create_experiment(type='FEATURE_FLAG')

        workspace = Mock(spec=Workspace)
        workspace.get_feature_flag_or_none.return_value = feature_flag
        self.workspace_fetcher.fetch.return_value = workspace

        evaluation = ExperimentEvaluation(
            DecisionReason.TRAFFIC_ALLOCATED,
            [],
            feature_flag,
            320,
            'A',
            None
        )
        self.experiment_evaluator.evaluate.return_value = evaluation
        self.event_factory.create.return_value = [Mock(), Mock()]

        # when
        actual = self.sut.feature_flag(42, self.user)

        # then
        self.assertEqual({}, actual.config)

    def test__track__event_Workspace를_가져오지_못하면_Undefined_이벤트를_전송한다(self):
        # given
        event = HackleEvent.builder('event_key').build()
        user = InternalHackleUser(identifiers={'$id': 'user_id'}, properties={})

        self.workspace_fetcher.fetch.return_value = None

        # when
        self.sut.track(event, user)

        # then
        self.event_processor.process.assert_called_once()

        track_event = self.event_processor.process.call_args[0][0]
        self.assertEqual(0, track_event.event_type.id)
        self.assertEqual('event_key', track_event.event_type.key)

    def test__track__event_eventKey에_대한_eventType을_찾지_못하면_Undeified_이벤트를_전송한다(self):
        # given
        event = HackleEvent.builder('event_key').build()
        user = InternalHackleUser(identifiers={'$id': 'user_id'}, properties={})

        workspace = Mock(spec=Workspace)
        workspace.get_event_type_or_none.return_value = None
        self.workspace_fetcher.fetch.return_value = workspace

        # when
        self.sut.track(event, user)

        # then
        self.event_processor.process.assert_called_once()

        track_event = self.event_processor.process.call_args[0][0]
        self.assertEqual(0, track_event.event_type.id)
        self.assertEqual('event_key', track_event.event_type.key)

    def test_track_event_설정정보에서_입력받은_이벤트키에_해당하는_이벤트타입정보로_이벤트를_전송한다(self):
        # given
        event = HackleEvent.builder('event_key').build()
        user = InternalHackleUser(identifiers={'$id': 'user_id'}, properties={})

        workspace = Mock(spec=Workspace)
        event_type = EventType(320, 'event_key')
        workspace.get_event_type_or_none.return_value = event_type
        self.workspace_fetcher.fetch.return_value = workspace

        # when
        self.sut.track(event, user)

        # then
        self.event_processor.process.assert_called_once()

        track_event = self.event_processor.process.call_args[0][0]
        self.assertEqual(320, track_event.event_type.id)
        self.assertEqual('event_key', track_event.event_type.key)

    def test_remote_config_설정정보를_가져오지_못하면_SDK_NOT_READY_사유와_default_value를_반환한다(self):
        # given
        user = InternalHackleUser(identifiers={'$id': 'user_id'}, properties={})
        self.workspace_fetcher.fetch.return_value = None

        # when
        decision = self.sut.remote_config('test_key', user, 'STRING', 'default')

        # then
        self.assertEqual(decision.reason, DecisionReason.SDK_NOK_READY)
        self.assertEqual(decision.value, 'default')

    def test_remote_config_parameter를_가져오지_못하면_REMOTE_CONFIG_PARAMETER_NOT_FOUND_사유와_default_value를_반환한다(self):
        # given
        user = InternalHackleUser(identifiers={'$id': 'user_id'}, properties={})
        workspace = Mock(spec=Workspace)
        self.workspace_fetcher.fetch.return_value = workspace
        workspace.get_remote_config_parameter_or_none.return_value = None

        # when
        decision = self.sut.remote_config('test_key', user, 'STRING', 'default')

        # then
        self.assertEqual(decision.reason, DecisionReason.REMOTE_CONFIG_PARAMETER_NOT_FOUND)
        self.assertEqual(decision.value, 'default')

    def test_remote_config_parameter를_정상적으로_가져오는_경우_해당_파라미터의_value_와_reason을_반환함(self):
        # given
        parameter = create_parameter()

        workspace = Mock(spec=Workspace)
        self.workspace_fetcher.fetch.return_value = workspace
        workspace.get_remote_config_parameter_or_none.return_value = parameter

        evaluation = RemoteConfigEvaluation(
            DecisionReason.DEFAULT_RULE,
            [],
            parameter,
            42,
            "vvv",
            {}
        )

        self.remote_config_evaluator.evaluate.return_value = evaluation
        self.event_factory.create.return_value = [Mock(), Mock()]

        # when
        decision = self.sut.remote_config('test_key', self.user, 'STRING', 'default')

        # then
        self.assertEqual(DecisionReason.DEFAULT_RULE, decision.reason)
        self.assertEqual('vvv', decision.value)
        self.assertEqual(2, self.event_processor.process.call_count)



from hackle.logger import adapt_logger, NoOpLogger


class Log(object):
    __GLOBAL_LOGGER = NoOpLogger().logger

    @staticmethod
    def initialize(logger):
        if logger is not None:
            Log.__GLOBAL_LOGGER = adapt_logger(logger)

    @staticmethod
    def get():
        """
        :rtype: logging.Logger 
        """
        return Log.__GLOBAL_LOGGER



import unittest
from unittest.mock import Mock

from hackle.internal.evaluation.action.action_resolver import ActionResolver
from hackle.internal.evaluation.bucket.bucketer import Bucketer
from hackle.internal.model.entities import TargetAction, Variation, Experiment, Bucket, Slot
from hackle.internal.user.internal_hackle_user import InternalHackleUser
from hackle.internal.workspace.workspace import Workspace
from tests.internal.evaluation.evaluator.experiment.test_experiment_request import create_experiment_request
from tests.internal.model.test_entities import create_experiment


class ActionResolverTest(unittest.TestCase):

    def setUp(self):
        self.bucketer = Mock(spec=Bucketer)
        self.sut = ActionResolver(self.bucketer)

    def test__unsupported_type(self):
        action = TargetAction(type='__INVALID__')

        actual = self.sut.resolve_or_none(create_experiment_request(), action)

        self.assertIsNone(actual)

    def test__variation_action__variation_id_에_해당하는_variation_을_가져온다(self):
        # given
        action = TargetAction(type='VARIATION', variation_id=420, bucket_id=None)
        variation = Mock(spec=Variation)
        experiment = Mock(spec=Experiment)
        experiment.id = 42
        experiment.get_variation_by_id_or_none.return_value = variation

        request = create_experiment_request(experiment=experiment)

        # when
        actual = self.sut.resolve_or_none(request, action)

        # then
        self.assertIsNotNone(actual)
        self.assertEqual(variation, actual)

    def test__variation_action__variation_id에_해당하는_Variation이_없으면_예외발생(self):
        # given
        action = TargetAction(type='VARIATION', variation_id=420)
        experiment = Mock(spec=Experiment)
        experiment.id = 42
        experiment.get_variation_by_id_or_none.return_value = None

        request = create_experiment_request(experiment=experiment)

        # when
        self.assertRaises(Exception, self.sut.resolve_or_none, request, action)

    def test__bucket_action__bucket_id에_해당하는_Bucket이_없으면_예외발생(self):
        # given
        action = TargetAction(type='BUCKET', bucket_id=320)
        workspace = Mock(spec=Workspace)
        workspace.get_bucket_or_none.return_value = None

        request = create_experiment_request(workspace=workspace)

        # when
        self.assertRaises(Exception, self.sut.resolve_or_none, request, action)

    def test__bucket_action__identifier_type에_해당하는_식별자가_없으면_None_리턴(self):
        # given
        user = InternalHackleUser(identifiers={'$id': 'test'}, properties={})
        action = TargetAction(type='BUCKET', bucket_id=320)
        bucket = Mock(spec=Bucket)
        experiment = create_experiment(identifier_type="custom_id")
        workspace = Mock(spec=Workspace)
        workspace.get_bucket_or_none.return_value = bucket

        request = create_experiment_request(workspace, user, experiment)

        # when
        actual = self.sut.resolve_or_none(request, action)

        # then
        self.assertIsNone(actual)

    def test__bucket_action__슬롯에_할당_안된경우_None(self):
        # given
        user = InternalHackleUser(identifiers={'$id': 'test'}, properties={})
        action = TargetAction(type='BUCKET', bucket_id=320)
        bucket = Mock(spec=Bucket)
        experiment = create_experiment()
        workspace = Mock(spec=Workspace)
        workspace.get_bucket_or_none.return_value = bucket
        self.bucketer.bucketing.return_value = None

        request = create_experiment_request(workspace, user, experiment)

        # when
        actual = self.sut.resolve_or_none(request, action)

        # then
        self.assertIsNone(actual)
        self.bucketer.bucketing.assert_called_once_with(bucket, 'test')

    def test__bucket_action__슬롯에_할당되었지만_슬롯의_variation_id에_해당하는_Variation이_없으면_None리턴(self):
        # given
        user = InternalHackleUser(identifiers={'$id': 'test'}, properties={})
        action = TargetAction(type='BUCKET', bucket_id=320)
        bucket = Mock(spec=Bucket)
        workspace = Mock(spec=Workspace)
        workspace.get_bucket_or_none.return_value = bucket

        slot = Slot(0, 100, 520)
        self.bucketer.bucketing.return_value = slot

        experiment = Mock(spec=Experiment)
        experiment.id = 42
        experiment.identifier_type = '$id'
        experiment.get_variation_by_id_or_none.return_value = None

        request = create_experiment_request(workspace, user, experiment)

        # when
        actual = self.sut.resolve_or_none(request, action)

        # then
        self.assertIsNone(actual)
        self.bucketer.bucketing.assert_called_once_with(bucket, 'test')
        experiment.get_variation_by_id_or_none.assert_called_once_with(520)

    def test__bucket_action__버켓팅을_통해_할당된_Variation을_리턴한다(self):
        # given
        user = InternalHackleUser(identifiers={'$id': 'test'}, properties={})
        action = TargetAction(type='BUCKET', bucket_id=320)
        bucket = Mock(spec=Bucket)
        workspace = Mock(spec=Workspace)
        workspace.get_bucket_or_none.return_value = bucket

        slot = Slot(0, 100, 520)
        self.bucketer.bucketing.return_value = slot

        experiment = Mock(spec=Experiment)
        experiment.id = 42
        experiment.identifier_type = '$id'
        experiment.get_variation_by_id_or_none.return_value = Variation(42, 'E', False)

        request = create_experiment_request(workspace, user, experiment)

        # when
        actual = self.sut.resolve_or_none(request, action)

        # then
        self.assertEqual(Variation(42, 'E', False), actual)
        self.bucketer.bucketing.assert_called_once_with(bucket, 'test')
        experiment.get_variation_by_id_or_none.assert_called_once_with(520)



from unittest import TestCase

from hackle.internal.evaluation.match.user_condition_matcher import UserValueResolver
from hackle.internal.model.entities import TargetKey
from hackle.internal.user.internal_hackle_user import InternalHackleUser


class UserValueResolverTest(TestCase):

    def setUp(self):
        self.sut = UserValueResolver()

    def test_USER_ID(self):
        user = InternalHackleUser(
            identifiers={
                '$id': 'test_user_id',
                '$deviceId': 'test_device_id',
                'custom_id': 'test_custom_id'
            },
            properties={}
        )
        self.assertEqual('test_user_id', self.sut.resolve_or_none(user, TargetKey('USER_ID', '$id')))
        self.assertEqual('test_device_id', self.sut.resolve_or_none(user, TargetKey('USER_ID', '$deviceId')))
        self.assertEqual('test_custom_id', self.sut.resolve_or_none(user, TargetKey('USER_ID', 'custom_id')))

    def test_USER_PROPERTY(self):
        user = InternalHackleUser(identifiers={'$id': 'test_user_id'}, properties={'age': 42})
        actual = self.sut.resolve_or_none(user, TargetKey('USER_PROPERTY', 'age'))
        self.assertEqual(42, actual)

    def test_HACKLE_PROPERTY(self):
        user = InternalHackleUser(identifiers={'$id': 'test_user_id'}, properties={'age': 42})
        actual = self.sut.resolve_or_none(user, TargetKey('HACKLE_PROPERTY', 'platform'))
        self.assertIsNone(actual)

    def test_SEGMENT(self):
        user = InternalHackleUser(identifiers={'$id': 'test_user_id'}, properties={'age': 42})
        with self.assertRaises(Exception) as actual:
            self.sut.resolve_or_none(user, TargetKey('SEGMENT', 'SEGMENT'))
        self.assertEqual('Unsupported target_key.type [SEGMENT]', str(actual.exception))



from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.model.entities import RemoteConfigParameter
from hackle.internal.user.internal_hackle_user import InternalHackleUser
from hackle.internal.workspace.workspace import Workspace


class RemoteConfigRequest(Evaluator.Request):
    def __init__(self, workspace, user, parameter, required_type, default_value):
        """
        :param Workspace workspace:
        :param InternalHackleUser user:
        :param RemoteConfigParameter parameter:
        :param str required_type:
        :param object default_value:
        """
        super(RemoteConfigRequest, self).__init__(Evaluator.Key('REMOTE_CONFIG', parameter.id), workspace, user)
        self.__parameter = parameter
        self.__required_type = required_type
        self.__default_value = default_value

    @property
    def parameter(self):
        """
        :rtype: RemoteConfigParameter
        """
        return self.__parameter

    @property
    def required_type(self):
        """
        :rtype: str
        """
        return self.__required_type

    @property
    def default_value(self):
        """
        :rtype: object
        """
        return self.__default_value

    def __str__(self):
        return 'EvaluatorRequest(type=REMOTE_CONFIG, key={})'.format(self.__parameter.key)

    def __repr__(self):
        return self.__str__()



import math

from hackle.internal.type import hackle_types
from tests import base


class HackleTypesTest(base.BaseTest):
    def test_is_string(self):
        self.assertTrue(hackle_types.is_string('str'))
        self.assertTrue(hackle_types.is_string('42'))
        self.assertTrue(hackle_types.is_string(''))
        self.assertFalse(hackle_types.is_string(42))
        self.assertFalse(hackle_types.is_string(True))
        self.assertFalse(hackle_types.is_string({}))
        self.assertFalse(hackle_types.is_string(None))

    def test_is_empty_string(self):
        self.assertTrue(hackle_types.is_empty_string(''))
        self.assertFalse(hackle_types.is_empty_string(' '))
        self.assertFalse(hackle_types.is_empty_string('42'))
        self.assertFalse(hackle_types.is_empty_string(42))
        self.assertFalse(hackle_types.is_empty_string(True))
        self.assertFalse(hackle_types.is_empty_string({}))
        self.assertFalse(hackle_types.is_empty_string(None))

    def test_is_not_empty_string(self):
        self.assertFalse(hackle_types.is_not_empty_string(''))
        self.assertTrue(hackle_types.is_not_empty_string(' '))
        self.assertTrue(hackle_types.is_not_empty_string('42'))
        self.assertFalse(hackle_types.is_not_empty_string(42))
        self.assertFalse(hackle_types.is_not_empty_string(True))
        self.assertFalse(hackle_types.is_not_empty_string({}))
        self.assertFalse(hackle_types.is_not_empty_string(None))

    def test_is_number(self):
        self.assertTrue(hackle_types.is_number(42))
        self.assertTrue(hackle_types.is_number(42.0))
        self.assertTrue(hackle_types.is_number(42.42))
        self.assertTrue(hackle_types.is_number(-42))
        self.assertTrue(hackle_types.is_number(-42.0))
        self.assertTrue(hackle_types.is_number(-42.42))
        self.assertTrue(hackle_types.is_number(math.nan))
        self.assertTrue(hackle_types.is_number(math.inf))
        self.assertFalse(hackle_types.is_number('42'))
        self.assertFalse(hackle_types.is_number(False))
        self.assertFalse(hackle_types.is_number(True))
        self.assertFalse(hackle_types.is_number({}))
        self.assertFalse(hackle_types.is_number(None))

    def test_is_finite_number(self):
        self.assertTrue(hackle_types.is_finite_number(42))
        self.assertTrue(hackle_types.is_finite_number(42.0))
        self.assertTrue(hackle_types.is_finite_number(42.42))
        self.assertTrue(hackle_types.is_finite_number(-42))
        self.assertTrue(hackle_types.is_finite_number(-42.0))
        self.assertTrue(hackle_types.is_finite_number(-42.42))
        self.assertFalse(hackle_types.is_finite_number(math.nan))
        self.assertFalse(hackle_types.is_finite_number(math.inf))
        self.assertFalse(hackle_types.is_finite_number('42'))
        self.assertFalse(hackle_types.is_finite_number(False))
        self.assertFalse(hackle_types.is_finite_number(True))
        self.assertFalse(hackle_types.is_finite_number({}))
        self.assertFalse(hackle_types.is_finite_number(None))

    def test_is_positive_int(self):
        self.assertTrue(hackle_types.is_positive_int(42))
        self.assertFalse(hackle_types.is_positive_int(42.0))
        self.assertFalse(hackle_types.is_positive_int(42.42))
        self.assertFalse(hackle_types.is_positive_int(0))
        self.assertFalse(hackle_types.is_positive_int(-42))
        self.assertFalse(hackle_types.is_positive_int(-42.0))
        self.assertFalse(hackle_types.is_positive_int(-42.42))
        self.assertFalse(hackle_types.is_positive_int(math.nan))
        self.assertFalse(hackle_types.is_positive_int(math.inf))
        self.assertFalse(hackle_types.is_positive_int('42'))
        self.assertFalse(hackle_types.is_positive_int(False))
        self.assertFalse(hackle_types.is_positive_int(True))
        self.assertFalse(hackle_types.is_positive_int({}))
        self.assertFalse(hackle_types.is_positive_int(None))

    def test_is_bool(self):
        self.assertTrue(hackle_types.is_bool(True))
        self.assertTrue(hackle_types.is_bool(False))
        self.assertFalse(hackle_types.is_bool(1))
        self.assertFalse(hackle_types.is_bool(0))
        self.assertFalse(hackle_types.is_bool(-1))
        self.assertFalse(hackle_types.is_bool("True"))
        self.assertFalse(hackle_types.is_bool("False"))



import math
import numbers

from six import integer_types
from six import string_types


# String type

def is_string(value):
    return value is not None and isinstance(value, string_types)


def is_empty_string(value):
    return is_string(value) and len(value) == 0


def is_not_empty_string(value):
    return is_string(value) and len(value) > 0


# Number type

def is_number(value):
    return value is not None and not isinstance(value, bool) and isinstance(value, numbers.Number)


def is_finite_number(value):
    if not isinstance(value, (numbers.Integral, float)):
        # numbers.Integral instead of int to accommodate long integer in python 2
        return False

    if isinstance(value, bool):
        # bool is a subclass of int
        return False

    if isinstance(value, float):
        if math.isnan(value) or math.isinf(value):
            return False

    if abs(value) > (2 ** 53):
        return False

    return True


def is_positive_int(value):
    return is_number(value) and isinstance(value, integer_types) and value > 0


def as_int_or_none(value):
    try:
        return int(value)
    except ValueError:
        return None


# Bool

def is_bool(value):
    return value is not None and isinstance(value, bool)



from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.match.condition_matcher_factory import ConditionMatcherFactory
from hackle.internal.logger.log import Log
from hackle.internal.model.entities import Target, TargetCondition


class TargetMatcher(object):

    def __init__(self, condition_matcher_factory):
        """
        :param ConditionMatcherFactory condition_matcher_factory:
        """
        self.condition_matcher_factory = condition_matcher_factory

    def matches(self, request, context, target):
        """
        :param Evaluator.Request request:
        :param Evaluator.Context context:
        :param Target target:

        :rtype: bool
        """
        for condition in target.conditions:
            if not self.__matches(request, context, condition):
                return False
        return True

    def __matches(self, request, context, condition):
        """
        :param Evaluator.Request request:
        :param Evaluator.Context context:
        :param TargetCondition condition:

        :rtype: bool
        """
        condition_matcher = self.condition_matcher_factory.get_condition_matcher_or_none(condition.key.type)
        if condition_matcher is None:
            Log.get().debug('Unsupported type [{}]. Please use the latest version of sdk.'.format(condition.key.type))
            return False
        return condition_matcher.matches(request, context, condition)
